# rhpkg - a Python library for RPM Packagers
#
# Copyright (C) 2011 Red Hat Inc.
# Author(s): Jesse Keating <jkeating@redhat.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

import glob
import os
import re
import subprocess
import tempfile
from collections import namedtuple

import git
import rpm
import six.moves.configparser as ConfigParser
from six import iteritems
from six.moves.urllib.parse import urlparse

import koji
import koji.util
import pyrpkg
from pyrpkg.sources import MalformedLineError, SourcesFile
from pyrpkg.utils import cached_property
from rhpkg.lookaside import RedHatLookasideCache

from . import rpmdefines
from .bugzilla_service import BugzillaService
from .covscan_service import CovscanManager
from .distgit_service import DistgitService
from .errata_service import ErrataWizard
from .product_pages_service import ProductPagesService

# Global stuff, that could go into a config one day

LookasideConfig = namedtuple('LookasideCacheConfig', 'lookaside lookasidehash lookaside_cgi')


SSL_VERIFY_FAILED_HINT = '''
Note: Do you have Red Hat IT Root CA certificate installed on your machine?
See https://mojo.redhat.com/docs/DOC-999615#jive_content_id_Certificates'''


# Add a decorator to check for the older kernel branches
def _check_old_kernel(func):
    """Check to see if the repo being worked on is an old kernel branch

    Will make a call to 'make' in order to get the configs in place
    """

    def checky(self, *args, **kwargs):
        # First only work on the kernel
        if self.repo_name == 'kernel':
            # Check to see if the version is one we care about
            oldkerns = ['2.4.21', '2.6.9', '2.6.18']
            if self.ver in oldkerns:
                # We have to make the configs first
                cmd = ['make', '-f', 'Makefile.config', 'configs']
                self._run_command(cmd, cwd=self.path)
        return func(self, *args, **kwargs)
    return(checky)


def detect_checksum(sourcesfile):
    """Open the sources file and try to guess what hash is used in the
    lookaside. We already assume that there will only be one type in the
    file. If there is no content or the file does not exist (or does not
    contain valid lines), ``None`` is returned.
    """
    for entry_type in ('old', 'bsd'):
        try:
            sf = SourcesFile(sourcesfile, entry_type=entry_type)
        except MalformedLineError:
            continue
        if sf.entries:
            return sf.entries[0].hashtype
    return None


class DictValuesAdapter(object):
    """
    Wrap a dict so it looks like a optparse.Values (an object with
    attributes) which can be passed to koji.util.maven_opts().
    """
    def __init__(self, data):
        self._data = data

    def __getattr__(self, name):
        value = self._data.get(name)
        if value is not None:
            return value
        raise AttributeError(name)


class Commands(pyrpkg.Commands):

    def __init__(self, path, lookaside, lookasidehash, *args, **kwargs):
        # Check what hash is used in sources file and use that. If detection
        # fails, fall back to md5.
        if lookasidehash == 'md5':
            detected = detect_checksum(os.path.join(path, 'sources'))
            lookasidehash = detected or lookasidehash

        # Call the super class init
        super(Commands, self).__init__(path, lookaside, lookasidehash, *args, **kwargs)

        # New properties
        self._brew_path = None
        self.cgit_url_template = ''
        self._brew_service = None
        self._anon_brew_service = None
        self._distgit_service = None
        self.ldapserver = 'ldap.corp.redhat.com'

        # For MD5 we want to use the old format of source files, the BSD format
        # should only be used when configured for SHA512
        self.source_entry_type = 'bsd' if self.lookasidehash != 'md5' else 'old'

    # Properites
    @property
    def brew_path(self):
        """This property ensures the brew_path attribute"""

        if not self._brew_path:
            self.load_brew_path()
        return self._brew_path

    def load_brew_path(self):
        """This sets the brew_path attribute"""
        brew_config = koji.read_config('brew')
        pathinfo = koji.PathInfo(brew_config['topdir'])
        build = self.brew_service.get_build_info(self.nvr)
        if build is None:
            raise RuntimeError('{0} is not built in Brew yet.'.format(self.nvr))
        else:
            self._brew_path = pathinfo.build(build)

    @property
    def brew_service(self):
        if not self._brew_service:
            self.load_brew_service()
        return self._brew_service

    @property
    def anon_brew_service(self):
        if not self._anon_brew_service:
            self.load_brew_service(anon=True)
        return self._anon_brew_service

    def load_brew_service(self, anon=False):
        if anon:
            self._anon_brew_service = BrewService(self.anon_kojisession, self.log)
        else:
            self._brew_service = BrewService(self.kojisession, self.log)

    @property
    def distgit_service(self):
        if not self._distgit_service:
            self.load_distgit_service()
        return self._distgit_service

    def load_distgit_service(self):
        self._distgit_service = DistgitService(self.repo, self.cgit_url_template, log=self.log)

    def is_version_released(self, osver):
        """Query Product Page to see if a product version is released"""
        self.log.debug('Determine release status of \'%s\'.', osver)
        # This whole thing needs error checking
        # Split the number by major/minor[/appstream]
        match = re.match(r'(?:supp-|sjis-|oracle-java-)?rhel-(\d+)\.(\d+)(?:\.(\d+))?', osver)
        if not match:
            # We don't know what product this is. It could be RHEL Alt, but
            # let's not do the transformation
            #     rhel-alt-7.4 -> rhel-pegas-1.0
            self.log.debug('Unknown release %s, assuming unreleased', osver)
            return False
        major, minor, appstream = match.groups()
        if appstream == '0' and minor == '0' and major == '8':
            # just for rhel-8.0.0
            label = 'rhel-8-0'
        elif appstream is not None:
            # rhel-8.0.1, rhel-8.1.0 and higher
            label = 'rhel-%s-%s.%s' % (major, minor, appstream)
        else:
            label = 'rhel-%s-%s' % (major, minor)

        pp = ProductPagesService(self.pp_url, log=self.log)
        # Load the data from product pages. There are two approaches to determine whether label
        # is released.
        # 1) look for specific task in release schedule (See COMPOSE-3858)
        if major >= '8':
            try:
                res = pp.is_released_as_per_schedule_task(label)
            except Exception:
                self.log.debug('Error getting product information for \'%s\', skipping '
                               'to next test.' % osver)
            else:
                # if task is not found or there is other error, continue with 2nd (original)
                # approach
                if res['found']:
                    return res['result']
        # 2) use direct method to find out whether label is released based on 'GA' date
        try:
            res = pp.is_released_as_per_ga(label)
        except Exception:
            self.log.debug('Error getting product information for \'%s\', assuming it\'s '
                           'unreleased.' % osver)
        else:
            if res['found']:
                return res['result']
            else:
                self.log.debug('Product information for \'%s\' does\'n exist, assuming it\'s '
                               'unreleased.' % osver)
                return False

    def is_osver_released(self, osver):
        if hasattr(self, 'released'):
            return self.released
        else:
            return self.is_version_released(osver)

    def reconcile_osver(self, osver):
        """Reconcile OS version according to whether it is released

        Rules to reconcile OS version:

        * If the osver has a decimal, see if it's been released.
        * If it has been released, use the z-stream dist naming, .elX_Y
        * otherwise use a generic .elX.  Thank Marketing for this.
        """
        if '.' in osver:
            self.log.debug('Seeing if %s is released', osver)
            if not self.is_osver_released(osver):
                osver = osver.split('.')[0]
        return osver

    def load_rpmdefines(self):
        """Populate rpmdefines based on branch data"""

        # We subclass in rhpkg so that we can look and see if a product has
        # released or not, as that will change the dist value. That only happens
        # to package builds relative to RHEL.

        loaders = {
            # This regex should find the 'rhel-5' or 'rhel-6.2' parts of the
            # branch name.  There should only be one of those, and all branches
            # should end in one.
            r'rhel-ready': self.load_rpmdefines_for_rhel_build,
            r'rhel(?:-alt)?-\d(\.\d+)?(\.\d+)?': self.load_rpmdefines_for_rhel_build,
            r'epel-\d(\.\d+)?': self.load_rpmdefines_for_rhel_build,
            r'fedora-\d+': self.load_rpmdefines_for_fedora_build,
        }

        for branch_regex, load_method in iteritems(loaders):
            match = re.search(branch_regex, self.branch_merge)
            if not match:
                continue

            load_method(match.group())

            self._rpmdefines = ["--define '_sourcedir %s'" % self.path,
                                "--define '_specdir %s'" % self.path,
                                "--define '_builddir %s'" % self.path,
                                "--define '_srcrpmdir %s'" % self.path,
                                "--define '_rpmdir %s'" % self.path,
                                "--define 'dist .%s'" % self._disttag,
                                "--define '%s %s'" % (self._distvar, self._distval.split('_')[0]),
                                "--define '%s 1'" % self._disttag]

            # For Fedora package build, macro _distunset is set.
            distunset = getattr(self, '_distunset', None)
            if distunset is not None:
                self._rpmdefines.append("--eval '%%undefine %s'" % distunset,)

            # For macros in local build for EPEL to match macros in brew build
            if 'epel' in self.branch_merge:
                self._rpmdefines.append('--define "rhel %s"' % self._distval.split('_')[0])

            return

        raise pyrpkg.rpkgError('Could not find the base OS ver from branch name %s. '
                               'Consider using --release option' % self.branch_merge)

    def load_rpmdefines_for_fedora_build(self, osver):
        self._distvar, self._distval = osver.split('-')

        # try to fetch extra data from brew. If data is missing,
        # fallback to standard %disttag detection
        extra_info = self.anon_brew_service.get_extra_info(self.branch_merge)
        if extra_info.get("rpm_macro_dist") or extra_info.get("rhpkg_dist"):
            # if both records are set, rpm_macro_dist has higher priority
            self._disttag = extra_info.get("rpm_macro_dist") or extra_info.get("rhpkg_dist")
            self.mockconfig = 'fedora-%s-%s' % (self._distval, self.localarch)
            self._distunset = 'rhel'
            self.log.debug('Using %%dist tag gathered from Brew extra info: %s', self._disttag)
            return

        for branch_regex, disttag_fmt in rpmdefines.from_fedora_branch:
            if re.search(branch_regex, self.branch_merge):
                self._disttag = disttag_fmt % {'distval': self._distval}
                self.mockconfig = 'fedora-%s-%s' % (self._distval, self.localarch)
                self._distunset = 'rhel'
                break

        if self._disttag is None:
            raise pyrpkg.rpkgError('Unable to load RPM build defines from unknown '
                                   'Fedora branch %s' % self.branch_merge)

    def load_rpmdefines_for_rhel_build(self, osver):
        osver = self.reconcile_osver(osver)
        # exception - version 8.0.0 doesn't exist. See COMPOSE-3792
        osver = osver.replace('rhel-8.0.0', 'rhel-8.0')

        self._distvar, self._distval = osver.rsplit('-', 1)
        self._distval = self._distval.replace('.', '_')

        # try to fetch extra data from brew. If data is missing,
        # fallback to standard %disttag detection
        extra_info = self.anon_brew_service.get_extra_info(self.branch_merge)
        if extra_info.get("rpm_macro_dist") or extra_info.get("rhpkg_dist"):
            # if both records are set, rpm_macro_dist has higher priority
            self._disttag = extra_info.get("rpm_macro_dist") or extra_info.get("rhpkg_dist")
            self.log.debug('Using %%dist tag gathered from Brew extra info: %s', self._disttag)
            return

        if 'rhel-alt-' in osver:
            # Remove -alt from the distvar. RHEL Alt branches should still
            # define a macro called `rhel`.
            self._distvar = self._distvar.replace('rhel-alt', 'rhel')

        for branch_regex, disttag_fmt in rpmdefines.from_rhel_branch:
            if re.match(branch_regex, self.branch_merge):
                self._disttag = disttag_fmt % {'distval': self._distval}
                break

        default_to_el = self._disttag is None
        if default_to_el:
            fmt = 'el%s' if 'rhel-alt' not in osver else 'el%sa'
            self._disttag = fmt % self._distval

            # Always Ready Kernel / Always Ready RHEL
            if 'rhel-ready' in osver:
                self._disttag = 'elrdy'

    def check_target_with_release_plan(self, container_build=False):
        """
        Checks the target availability for specific RHEL version
        with Product Pages and modifies target acordingly if needed.
        """
        # This piece of code is for issue COMPOSE-359. For details of why this
        # piece of code works in this way, please refer to COMPOSE-359.
        # See COMPOSE-2681 and COMPOSE-3117 and COMPOSE-3399 for more details.
        branch_name = self.branch_merge
        omatch = re.match(
            r'^(?:supp-|sjis-|oracle-java-)?rhel-(\d+)\.(\d+)(?:\.(\d+))?$',
            branch_name)
        if omatch is not None:
            X, Y, Z = omatch.groups()
            # The user did not supply target explicitly, and we are on a
            # RHEL branch.
            if self.is_version_released(branch_name):
                # /Product Pages say the version corresponding to current
                # branch is released. That means the build should always go
                # into Z target.
                if container_build:
                    self._container_build_target = "%s-z-containers-candidate" % branch_name
                else:
                    self._target = '%s-z-candidate' % branch_name
            # version is not released and version is 7 and older.
            elif Z is None:
                # The current version is not released. We can check if a
                # branch for the next version exists. If not, current
                # version is in development and there is nothing special to
                # do.
                next_version = branch_name.replace(
                    "rhel-%s.%s" % (X, Y),
                    "rhel-%s.%s" % (X, int(Y) + 1),
                )
                has_next_version = self.repo.git.ls_remote(
                    '.', 'origin/%s' % next_version) != ''
                if has_next_version:
                    # There is a branch for the next version. The build
                    # could be either for the Z stream of current version,
                    # or it's a fix for a blocker bug on the current
                    # release. No way to tell automatically, so ask user.
                    raise ValueError(
                        'Please supply proper --target argument. Target '
                        'should most likely be "%s-z-candidate" '
                        'unless building fixes of blocker bugs.' % branch_name)

    def load_target(self):
        # sed-rhel-* does not use -candidate
        if re.match(r'sed-rhel-\d', self.branch_merge):
            self._target = self.branch_merge
        else:
            super(Commands, self).load_target()

    def load_container_build_target(self):
        # TODO this will likely change as packages move to other namespaces in
        # dist-git. Ideally we would want to embed namespace into the target
        containers_tgt = '%s-containers-candidate' % (self.branch_merge)
        # Check if the -containers-candidate target exists, and if not, default
        # to the old -docker-candidate.
        target = self.kojisession.getBuildTarget(containers_tgt)
        if target:
            self._container_build_target = containers_tgt
        else:
            self._container_build_target = '%s-docker-candidate' % self.branch_merge

    def check_target(self, target, will_tag):
        """Check that the target exists.  If will_tag is true, also check that
        the target isn't locked."""
        # Check if the target is valid
        build_target = self.kojisession.getBuildTarget(target)
        if not build_target:
            raise pyrpkg.rpkgError('Unknown build target: %s' % target)
        # Check if the dest tag is locked
        dest_tag = self.kojisession.getTag(build_target['dest_tag_name'])
        if not dest_tag:
            raise pyrpkg.rpkgError('Unknown destination tag %s' % build_target['dest_tag_name'])
        if will_tag and dest_tag['locked']:
            raise pyrpkg.rpkgError('Destination tag %s is locked' % dest_tag['name'])

    def precheck(self, target, will_tag):
        """Run checks before launching the build.
        This will check that:
         - there are no uncommitted and/or unpushed changes in the local repo
         - the target exists
        If will_tag is True, it will also verify that the destination tag is
        not locked.
        XXX This really belongs in rpkg.
        """
        self.check_repo()
        self.check_target(target, will_tag)

    def resolve_relative_urls(self, opts, inis, section):
        """.ini files may contain:
        patches = .
          or
        specfile = .
        which means use the SCM URL of the directory that the .ini is in.
        Find which of the .ini files contained the specified section, and
        replace any '.' entries with the SCM URL of that .ini.  If the .ini
        is not in a properly formatted repository an error will be raised.
        """
        for ini in inis:
            with open(ini) as iobj:
                conf = ConfigParser.ConfigParser()
                conf.readfp(iobj)
                if not section:
                    # only happens when we have exactly one .ini with
                    # exactly one section
                    section = conf.sections()[0]
                if conf.has_section(section):
                    # we need to find the giturl for a different directory,
                    # so temporarily change the path attribute
                    try:
                        origpath = self.path
                        self.path = os.path.dirname(ini)
                        # clear some state that should be cleared by
                        # setting self.path, but isn't
                        self._repo_name = None
                        self._commit = None
                        giturl = self.giturl()
                    finally:
                        self.path = origpath
                        self._repo_name = None
                        self._commit = None
                    for key in ['patches', 'specfile']:
                        if opts.get(key) == '.':
                            self.log.debug('%s: replacing %s = . with %s', ini, key, giturl)
                            opts[key] = giturl
                    return

    # Commands
    def maven_build(self, sources, patches=None, goals=None, profiles=None,
                    properties=None, envs=None, packages=None, jvm_options=None,
                    maven_options=None, inis=None, section=None, specfile=None,
                    debug=False, background=False, skip_tag=False, scratch=False,
                    target_override=False):
        """Initiate a maven-build of the package.

        sources: The URL to the sources to build
        patches: URL to the patches, will be discovered if not supplied
        goals: Optional list of goals to pass on to Maven
        profiles: Optional list of profiles to pass to Maven
        properties: Optional list of name=value or name properties to pass
                    to Maven
        envs: Optional list of environment variables to set
        packages: Optional list of extra packages to install in the buildroot
        jvm_options: Optional list of options to pass to the JVM
        maven_options: Optional list of options to pass to Maven
        specfile: URL to the .spec.tmpl, will be discovered if not supplied
        debug: Run Maven in debug mode
        background: Perform the build with a low priority
        skip_tag: Do not attempt to tag the build
        scratch: Perform a scratch build
        """

        # Start building up the command
        cmd = [self.build_client]
        maven_target = self.target
        if not target_override:
            # Translate the build target into a maven target,
            # but only if --target wasn't specified on the command-line
            maven_target = '%s-maven-candidate' % self.target.split('-candidate')[0]
        # sanity check
        self.precheck(maven_target, not (scratch or skip_tag))
        cmd.append('maven-build')
        # define our dictionary for options
        if inis:
            for ini in inis:
                cmd.extend(['--ini', ini])
            if section:
                cmd.extend(['--section', section])
            try:
                params = koji.util.parse_maven_param(inis, section=section, scratch=scratch)
                opts = list(params.values())[0]
                sources = opts.pop('scmurl')
            except ValueError as e:
                raise pyrpkg.rpkgError(*e.args)
            if opts.get('patches') == '.' or opts.get('specfile') == '.':
                self.resolve_relative_urls(opts, inis, section)
        else:
            opts = koji.util.maven_opts(DictValuesAdapter(locals()),
                                        scratch=scratch)

            # Discover items
            # patches
            if patches:
                if patches == '.':
                    patches = self.giturl()
                    opts['patches'] = patches
                cmd.extend(['--patches', patches])
            # specfile
            if specfile:
                if specfile == '.':
                    specfile = self.giturl()
                    opts['specfile'] = specfile
                cmd.extend(['--specfile', specfile])

            # Reflect all the build options in the cmd list
            # This seems really redundant
            for goal in opts.get('goals', []):
                cmd.extend(['-G', goal])
            for prof in opts.get('profiles', []):
                cmd.extend(['-P', prof])
            for name, value in opts.get('properties', {}).items():
                if value is None:
                    cmd.extend(['-D', name])
                else:
                    cmd.extend(['-D', '%s=%s' % (name, value)])
            for name, value in opts.get('envs', {}).items():
                cmd.extend(['-E', '%s=%s' % (name, value)])
            for pkg in opts.get('packages', []):
                cmd.extend(['-p', pkg])
            for opt in opts.get('jvm_options', []):
                cmd.extend(['-J', opt])
            for opt in opts.get('maven_options', []):
                cmd.extend(['-M', opt])

        if debug:
            cmd.append('--debug')
            opts.setdefault('maven_options', []).append('--debug')
        # Set a placeholder for the build priority
        priority = None
        if background:
            cmd.append('--background')
            priority = 5  # magic koji number :/
        if scratch:
            cmd.append('--scratch')
        elif skip_tag:
            cmd.append('--skip-tag')
            opts['skip_tag'] = True
        # This target should be a maven target
        cmd.append(maven_target)

        # Now send off the build
        cmd.append(sources)
        self.log.info('Building %s in %s', sources, maven_target)
        self.log.debug('Building %s in %s with options %s and a priority of %s',
                       sources, maven_target, opts, priority)
        self.log.debug(' '.join(cmd))
        task_id = self.kojisession.mavenBuild(sources, maven_target, opts, priority=priority)
        self.log.info('Created task: %s' % task_id)
        self.log.info('Task info: %s/taskinfo?taskID=%s', self.kojiweburl, task_id)
        return task_id

    def maven_chain(self, inis,
                    debug=False, background=False, skip_tag=False, scratch=False,
                    target_override=False):
        """Initiate a Maven chain-build.

        inis: A list of .ini files to load build parameters from
        debug: Run Maven in debug mode
        background: Perform the build with a low priority
        skip_tag: Do not attempt to tag the build
        scratch: Perform a scratch build
        """

        # Start building up the command
        cmd = [self.build_client]
        maven_target = self.target
        if not target_override:
            # Translate the build target into a maven target,
            # but only if --target wasn't specified on the command-line
            maven_target = '%s-maven-candidate' % self.target.split('-candidate')[0]
        # sanity check
        self.precheck(maven_target, not (scratch or skip_tag))
        cmd.append('maven-chain')
        builds = koji.util.parse_maven_chain(inis, scratch=scratch)
        for pkgname, params in builds.items():
            if params.get('patches') == '.' or params.get('specfile') == '.':
                self.resolve_relative_urls(params, inis, pkgname)
        opts = {}
        if debug:
            cmd.append('--debug')
            opts['debug'] = True
        # Set a placeholder for the build priority
        priority = None
        if background:
            cmd.append('--background')
            priority = 5  # magic koji number :/
        if scratch:
            cmd.append('--scratch')
            opts['scratch'] = True
        elif skip_tag:
            cmd.append('--skip-tag')
            opts['skip_tag'] = True
        # This target should be a maven target
        cmd.append(maven_target)

        # Now send off the build
        cmd.extend(inis)
        self.log.info('Building %s in %s', ', '.join(builds.keys()), maven_target)
        self.log.debug('Building %s in %s with options %s and a priority of %s',
                       ', '.join(builds.keys()), maven_target, opts, priority)
        self.log.debug(' '.join(cmd))
        task_id = self.kojisession.chainMaven(builds, maven_target, opts, priority=priority)
        self.log.info('Created task: %s' % task_id)
        self.log.info('Task info: %s/taskinfo?taskID=%s', self.kojiweburl, task_id)
        return task_id

    def wrapper_rpm(self, build, specfile, create_build=False,
                    background=False, skip_tag=False, scratch=False):
        """Initiate a wrapper-rpm task.

        build: The NVR or ID of the build to create wrapper rpms for
        specfile: URL of the .spec.tmpl
        create_build: Create a new build and associate the wrapper rpms
                      with it, rather than associating them with the
                      existing build
        skip_tag: Do not attempt to tag the build
        scratch: Perform a scratch build
        """

        # Start building up the command
        cmd = [self.build_client]
        # See if repo is dirty
        if self.repo.is_dirty():
            raise pyrpkg.rpkgError('There are uncommitted changes in your repo')
        # Need to check here to see if the local commit you want to build is
        # pushed or not
        branch = self.repo.active_branch
        branch_merge = self.branch_merge
        if self.repo.git.rev_list('%s...%s' % (branch, branch_merge)):
            raise pyrpkg.rpkgError('There are unpushed changes in your repo')
        # Check to see if the target is valid
        build_target = self.kojisession.getBuildTarget(self.target)
        if not build_target:
            raise pyrpkg.rpkgError('Unknown build target: %s' % self.target)
        # see if the dest tag is locked
        dest_tag = self.kojisession.getTag(build_target['dest_tag_name'])
        if not dest_tag:
            raise pyrpkg.rpkgError('Unknown destination tag %s' % build_target['dest_tag_name'])
        if dest_tag['locked'] and create_build and not (scratch or skip_tag):
            raise pyrpkg.rpkgError('Destination tag %s is locked' % dest_tag['name'])
        cmd.append('wrapper-rpm')
        # Set a placeholder for the build priority
        priority = None
        if background:
            cmd.append('--background')
            priority = 5  # magic koji number :/
        opts = {}
        if scratch:
            cmd.append('--scratch')
            opts['scratch'] = True
        elif create_build:
            cmd.append('--create-build')
            opts['create_build'] = True
            if skip_tag:
                cmd.append('--skip-tag')
                opts['skip_tag'] = True
        cmd.append(self.target)
        if build.isdigit():
            build = int(build)
        cmd.append(build)
        if specfile == '.':
            specfile = self.giturl()
        cmd.append(specfile)
        self.log.info('Building wrapper rpms for %s in %s', build, self.target)
        self.log.debug('Building wrapper rpms for %s in %s with options %s and a priority of %s',
                       build, self.target, opts, priority)
        self.log.debug(' '.join(cmd))
        task_id = self.kojisession.wrapperRPM(build, specfile, self.target,
                                              priority=priority, opts=opts)
        self.log.info('Created task: %s', task_id)
        self.log.info('Task info: %s/taskinfo?taskID=%s', self.kojiweburl, task_id)
        return task_id

    def tests(self, module, path=None, branch=None, anon=False):
        """Clone the tests repo, optionally check out a specific branch.

        module is the name of the module to clone
        path is the basedir to perform the clone in
        branch is the name of a branch to checkout instead of <remote>/master
        anon is whether or not to clone anonymously
        Logs the output and returns nothing.

        """

        if not path:
            path = self.path
        # construct the git url
        if anon:
            giturl = self._get_namespace_anongiturl(module)
        else:
            giturl = self._get_namespace_giturl(module)
        giturl = giturl.replace('rpms', 'tests')

        # Create the command
        cmd = ['git', 'clone']
        if self.quiet:
            cmd.append('--quiet')
        # do the clone
        if branch:
            self.log.debug('Checking out a specific branch %s', giturl)
            cmd.extend(['-b', branch, giturl])
        else:
            self.log.debug('Cloning %s', giturl)
            cmd.extend([giturl])
        self._run_command(cmd, cwd=path)

        return

    def publican_package(self, lang, config, publican_args=None):
        """Invoke publican to create an srpm of a docs package

        lang is the language of the doc (or "none" if no language)
        config is the path to the publican configuration file
        publican_args is a string of additional arguments to parse to publican

        """

        # check publican is installed (using the rpm database)
        pub_rpms = ["%s-%s-%s" % (h["name"], h["version"], h["release"])
                    for h in rpm.TransactionSet().dbMatch('name', 'publican')]
        if pub_rpms == []:
            raise Exception('Publican is not installed.')

        # construct the command line
        cmd = ["publican", "package", "--config", config]
        if lang != "none":
            cmd += ["--lang", lang]
        if publican_args is not None:
            cmd.extend(publican_args.split(" "))

        # run publican
        self._run_command(cmd)

        # find the generated srpm
        srpm = glob.glob("tmp/rpm/*.src.rpm")
        if len(srpm) == 0:
            raise Exception('Publican failed to produce an SRPM')
        elif len(srpm) > 1:
            raise Exception('Publican appears to have produced multiple SRPMS')

        return srpm[0]

    def publican_git_commit(self, srpm, branch, message):
        """Update dist-git with publican generated srpm

        srpm is the path to the publican generated srpm
        branch is the git branch to be used
        message is the commit message to be used
        """

        # work out the cwd to use absolute paths
        base_dir = os.getcwd()

        # work out the package name
        pkgname = os.path.basename(srpm).rsplit("-", 2)[0]

        # get a clone of the dist-git repo to commit to
        clone_dir = "%s/tmp/dist-git/%s" % (base_dir, pkgname)
        if not os.path.exists(clone_dir):
            self.log.debug('Cloning %s into tmp/dist-git', pkgname)
            if not os.path.exists("tmp/dist-git"):
                os.makedirs("tmp/dist-git")
            self.clone(pkgname, path="tmp/dist-git", anon=False, branch=branch)
            self.path = clone_dir
        else:
            self.log.debug('Using pre-existing clone in tmp/dist-git/%s', pkgname)
            self.path = clone_dir
            self.switch_branch(branch)
            self.pull()

        if self.branch_merge != branch:
            raise Exception("Could not switch to branch '%s'" % branch)

        # import the srpm
        self.log.info('Importing publican generated srpm')
        upload_files = self.import_srpm(os.path.join(base_dir, srpm))
        # We get only file names here. We know though that the files are in the
        # dist-git checkout.
        upload_files = [os.path.join(clone_dir, f) for f in upload_files]

        # upload the base files
        self.log.info('Uploading files to lookaside cache')
        self.upload(upload_files, replace=True)

        # do the commit and push it to dist-git
        self.log.info('Committing and pushing changes to dist-git')
        self.commit(message)
        self.push()

    def covscan(self, srpm_path, mock_profile=None, covscan_args=None, base_srpm_path=None,
                latest=None, check_patches=None, executable='covscan'):
        """
        Initiate static analysis scan in covscan tool.

        covscan_args is a string with additional arguments for covscan
        """
        branch = self.branch_merge
        cm = CovscanManager(executable, branch, mock_profile=mock_profile, log=self.log,
                            covscan_args=covscan_args)
        if check_patches:
            return cm.diff_analyze(srpm_path)

        # base = None -- latest; base = "" -- not specified
        if not base_srpm_path and not latest:
            return cm.analyze(srpm_path)
        else:
            return cm.version_diff_analyze(srpm_path, base_srpm_path=base_srpm_path,
                                           base_brew_build=latest)

    def rpmdiff(self, hub_url, scratch_id, base=None):
        """Schedule RPMDiff run

        :param str hub_url: the RPMDiff Hub in where to schedule run
        :param int scratch_id: task ID of a scratch build that will be analyzed
            or compared with baseline
        :param str base: optional. The baseline with which to compare the scratch build.
            Analysis run will be scheduled if base is not provided.
        """
        # verify whether 'rpmdiff-remote' command is available. Otherwise do nothing.
        ret, _, _ = self._run_command(["which", "rpmdiff-remote"], shell=True, return_stderr=True)
        if ret != 0:
            self.log.error("'rpmdiff-remote' command is not available on current environment")
            return

        if base is None:
            cmd = ['rpmdiff-remote', '--hub-url', hub_url, 'schedule', str(scratch_id)]
        else:
            cmd = ['rpmdiff-remote', '--hub-url', hub_url, 'schedule', '-b', base, str(scratch_id)]
        self._run_command(cmd)

    def errata(self, et_conf, umb_config, erratum=None,
               product_versions=None, bugs_to_add=None, bugs_to_remove=None,
               builds=None, move_to_newfiles=None, move_to_qe=None,
               dry_run=False, debug=False, verbose=False, username=None,
               password=None, message=None, nowait_rpmdiff=False,
               bugzilla_url=None, bugzilla_username=None, bugzilla_password=None):
        """
        et_conf [dict]    -- configuration of errata tool containing base address
        umb_config [dict] -- configuration of message bus where rpmdiff sends updates

        Do all the required operations with erratum
        """
        em = ErrataWizard(
            et_conf,
            umb_config,
            package=self.repo_name,
            erratum=erratum,
            product_versions=product_versions,
            bugs_to_add=bugs_to_add,
            bugs_to_remove=bugs_to_remove,
            builds=builds,
            move_to_newfiles=move_to_newfiles,
            move_to_qe=move_to_qe,
            dry_run=dry_run,
            log=self.log,
            debug=debug,
            verbose=verbose,
            username=username,
            password=password,
            message=message,
            bugzilla_url=bugzilla_url,
            bugzilla_username=bugzilla_username,
            bugzilla_password=bugzilla_password,
            nowait_rpmdiff=nowait_rpmdiff,
        )
        em.run()

    def bugzilla(self, url, bug_id, username=None, password=None, fixed_in_nvrs=None,
                 to_modified=False, commit_urls=None, dry_run=False, debug=False, verbose=False):
        """
        et_conf [string]          -- configuration of errata tool -- base address
        rpmdiff_msg_conf [string] -- configuration of message bus where rpmdiff sends updates

        Do all the required operations with erratum
        """
        bz = BugzillaService(url, username=username, password=password, log=self.log,
                             dry_run=dry_run)
        bz.batch_update(bug_id,
                        fixed_in_nvrs=fixed_in_nvrs,
                        to_modified=to_modified,
                        commit_urls=commit_urls)

    # Overloaded commands
    @_check_old_kernel
    def compile(self, *args, **kwargs):
        # Call the super class
        super(Commands, self).compile(*args, **kwargs)

    @_check_old_kernel
    def install(self, *args, **kwargs):
        # Call the super class
        super(Commands, self).install(*args, **kwargs)

    @_check_old_kernel
    def local(self, *args, **kwargs):
        # Call the super class
        super(Commands, self).local(*args, **kwargs)

    @_check_old_kernel
    def prep(self, *args, **kwargs):
        # Call the super class
        super(Commands, self).prep(*args, **kwargs)

    @_check_old_kernel
    def srpm(self, *args, **kwargs):
        # Call the super class
        super(Commands, self).srpm(*args, **kwargs)

    def find_remote_backend_config_file(self, remote_backend):
        """Find remote configuration file

        Generally, *pkg configuration file should be put under /etc/rpkg/,
        this is the first place to find. If there is no such a configuration
        file, for example fedpkg.conf, that usually means fedpkg could not be
        installed yet. In this case, FEDPKG_CONFIG environment variable will be
        checked. For centpkg, it's CENTPKG_CONF.

        If configure file cannot be found from any place above, an error is
        raised.
        """
        config_file = '/etc/rpkg/{0}.conf'.format(remote_backend)
        if not os.path.exists(config_file):
            env_var = '{0}_CONFIG'.format(remote_backend.upper())
            config_file = os.environ.get(env_var)
            if config_file is None or not os.path.exists(config_file):
                raise RuntimeError(
                    'Cannot find configuration file: {0}. '
                    'Please ensure to install {1}.'.format(config_file, remote_backend))
        return config_file

    def remote(self, remote_name, repo_name, remote_backends):
        """Add remote pointing repository from remote dist-git service

        fedpkg.conf or centpkg.conf will be discovered automatically. If it
        cannot be found, an exception will be raised to notice user to install
        fedpkg or centpkg.

        Read-only remote is added in case something is pushed back
        accidentally.

        There is no any magic by adding a remote here. It's just wrapper of
        git-remote.

        :param str remote_name: remote to add
        :param str repo_name: remote repository name. If remote repository has
        same name as current repository, repo_name can be None
        :param dict remote_backends: a mapping from remote name to backend name
        """
        backend_cli = remote_backends['{0}_cli'.format(remote_name)]
        remote_config_file = self.find_remote_backend_config_file(backend_cli)
        config = ConfigParser.RawConfigParser()
        self.log.debug('Reading configuration from %s: %s', remote_name, remote_config_file)
        config.read(remote_config_file)

        module_name = repo_name or self.ns_repo_name

        anongiturl = config.get(backend_cli, 'anongiturl') % {
            'repo': module_name,
            'module': module_name,
        }
        self._run_command(['git', 'remote', 'add', remote_name, anongiturl])
        self._run_command(['git', 'fetch', remote_name])

    def find_local_sources(self):
        """Find local sources file"""
        filename = os.path.join(os.path.abspath(os.curdir), 'sources')
        self.log.debug('Finding local sources file: %s', filename)
        if not os.path.exists(filename):
            raise RuntimeError('Cannot find sources file from current directory.')
        return filename

    def find_remote_sources_file(self, remote, branch):
        remote_branch = '{0}/{1}:sources'.format(remote, branch)
        cmd = ['git', 'cat-file', '-p', remote_branch]
        self.log.debug(cmd)
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()

        if proc.returncode > 0:
            raise RuntimeError('Fail to get sources files from {0}'.format(remote_branch))

        fd, filename = tempfile.mkstemp()
        os.write(fd, stdout)
        os.close(fd)

        return filename

    def read_lookaside_config(self, config_file, remote_backend):
        """Read lookasidecache configuration file from backend config file"""
        config = ConfigParser.RawConfigParser()
        config.read(config_file)
        return LookasideConfig(
            lookasidehash=config.get(remote_backend, 'lookasidehash'),
            lookaside=config.get(remote_backend, 'lookaside'),
            lookaside_cgi=config.get(remote_backend, 'lookaside_cgi'))

    @cached_property
    def lookasidecache(self):
        """A helper to interact with the lookaside cache

        If we use MD5 hash, we can use the same path that default rpkg is
        using, otherwise we need to override it to make sure the hash type is
        included in the URL.
        """
        if self.lookasidehash != 'md5':
            return RedHatLookasideCache(
                self.lookasidehash, self.lookaside, self.lookaside_cgi)
        return super(Commands, self).lookasidecache

    def get_repo_name(self, remote=None):
        if remote is None:
            return self.repo_name
        else:
            url = self.repo.git.config('remote.{0}.url'.format(remote))
            return os.path.basename(urlparse(url).path).replace('.git', '')

    def download_sources_files(self, sources_filename, lookaside_config, repo_name, outdir):
        """Download files listed in sources file

        :param str sources_filename: filename of sources file that contains
        file to download.
        :param LookasideCacheConfig lookaside_config: object containing
        lookaside cache configuration that can be used to download files.
        """
        sourcesf = SourcesFile(sources_filename, self.source_entry_type)

        # Currently, no need to pass cert info to lookaside. Here is the
        # right place to handle and pass cert info to lookaside if it's
        # necessary
        lookasidecache = RedHatLookasideCache(lookaside_config.lookasidehash,
                                              lookaside_config.lookaside,
                                              lookaside_config.lookaside_cgi)

        def download_source_file(entry):
            outfile = os.path.join(outdir, entry.file)
            lookasidecache.download(repo_name,
                                    entry.file,
                                    entry.hash,
                                    outfile,
                                    hashtype=entry.hashtype)

        for entry in sourcesf.entries:
            download_source_file(entry)

    def get_referenced_dist_git_netloc(self, remote):
        url = self.repo.git.config('--get', 'remote.{0}.url'.format(remote))
        urlparts = urlparse(url)
        if len(urlparts.netloc) == 0:
            if os.path.exists(urlparts.path):
                raise RuntimeError('Guessing you are working on a repo cloned'
                                   ' from local filesystem, that is not supported now.')
            else:
                raise RuntimeError('Cannot get domain from URL {0}. '
                                   'Make sure you are working on a valid repo.'.format(url))
        # We don't need the username part
        return urlparts.netloc.split('@')[-1]

    def is_remote_referencing_same_dist_git(self, remote_name, baseline_remote=None):
        _baseline_remote = baseline_remote or 'origin'
        if remote_name == _baseline_remote:
            return True
        dg_netloc = self.get_referenced_dist_git_netloc(remote_name)
        baseline_netloc = self.get_referenced_dist_git_netloc(_baseline_remote)
        return dg_netloc == baseline_netloc

    def get_remote_netloc(self, remote_name):
        url = self.repo.git.config('--get', 'remote.{0}.url'.format(remote_name))
        return urlparse(url).netloc.split('@')[-1]

    def exploded_sources(self):
        """Download exploded sources listed in source-repo file"""
        source_repos_file = os.path.join(self.path, 'source-repos')
        if not os.path.exists(source_repos_file):
            return

        ore = re.compile(r'^([\S]+)\s+([\S]+)$')

        def read_exploded_sources():
            self.log.debug('Read exploded sources from %s', source_repos_file)
            with open(source_repos_file, 'r') as f:
                for line in f:
                    source_line = line.strip()
                    match = ore.match(source_line)
                    if not match:
                        raise RuntimeError('Invalid line format in source-repos: '
                                           '{0}'.format(source_line))
                    else:
                        repourl, commit = match.groups()
                        if commit == 'HEAD':
                            raise RuntimeError('Invalid line format in source-repos: '
                                               'you may not use HEAD as commit ID.')
                        yield repourl, commit

        for repourl, commit in read_exploded_sources():
            self.log.debug('Make tarball from repo %s and commit %s',
                           repourl, commit)
            reponame = os.path.basename(repourl).replace('.git', '')
            repo_path = os.path.join(self.path, reponame)
            try:
                git.Repo(repo_path)
            except git.NoSuchPathError:
                self._run_command(['git', 'clone', '-n', repourl, reponame],
                                  cwd=self.path)

            tarball_file = os.path.join(
                self.path, '{0}-{1}.tar.gz'.format(reponame, commit))
            cmd = (
                'git archive --format=tar --prefix="{0}-{1}/" "{1}" | '
                'gzip -n -9 > "{2}"'.format(reponame, commit, tarball_file)
            )
            self.log.debug('Run: %s', cmd)
            # Return the exit code of the first command
            # PIPESTATUS is for bash, zero based array
            # pipestatus is for zsh, 1 based array
            cmd = '%s; exit "${PIPESTATUS[0]} ${pipestatus[1]}"' % cmd
            try:
                subprocess.check_call(cmd, shell=True, cwd=repo_path)
            except subprocess.CalledProcessError:
                if os.path.exists(tarball_file):
                    os.unlink(tarball_file)
                raise RuntimeError('Failed to create archive from commit {0}'.format(commit))

    def sources(self, outdir=None, remote=None, branch=None, remote_backends=None):
        """Download sources files in a more smart way

        Multiple remote may be added by remote command, that is, there would be
        another remote named fedora and pointing to a same module within Fedora
        dist-git. sources is smart enough to know which remote current branch
        is on and from where to download source files.

        If current branch is on origin remote, sources is able to download
        source files in the original way done by rpkg, and it accepts remote
        or both remote and branch to download from added remote (the repository
        within Fedora dist-git).

        If current branch is the one tracking remote branch. sources is able to
        download source files from corresponding remote repository
        automatically.

        Formats:

        rhpkg sources
        rhpkg sources fedora
        rhpkg sources fedora f25
        """
        if outdir is None:
            outdir = self.path

        current_remote = self.branch_remote

        if self.is_remote_referencing_same_dist_git(current_remote, 'origin'):
            if remote is None and branch is None:
                super(Commands, self).sources(outdir)
            else:
                backend_cli = remote_backends['{0}_cli'.format(remote)]
                config_file = self.find_remote_backend_config_file(backend_cli)
                lookaside_config = self.read_lookaside_config(config_file, backend_cli)
                repo_name = self.get_repo_name(remote)
                if branch:
                    sources_filename = self.find_remote_sources_file(remote, branch)
                else:
                    sources_filename = self.find_local_sources()
                self.download_sources_files(
                    sources_filename, lookaside_config, repo_name, outdir)
        else:
            backend_cli = None
            netloc = self.get_remote_netloc(current_remote)
            for name, value in iteritems(remote_backends):
                if name.endswith('_dist_git') and value == netloc:
                    backend_cli = remote_backends['{0}_cli'.format(name.split('_')[0])]
                    break
            if backend_cli is None:
                raise RuntimeError('Unsupported remote {0}'.format(current_remote))
            self._run_command([backend_cli, 'sources'])

        self.exploded_sources()

    def user_is_devel_member(self):
        """Gets 'devel' group memberships for active kerberos user
        :return: True if user is member of devel group, else False
        :rtype: bool
        """
        cmd = [  # group query for current active kerberos user
            'ldapsearch',
            '-LLL',
            '-x',
            '-h',
            self.ldapserver,
            '-b',
            'ou=groups,dc=redhat,dc=com',
            '(&(cn=devel)(memberUid=%s))' % self.user
        ]
        ret, stdout, stderr = self._run_command(cmd, return_stdout=True, return_stderr=True)
        if ret:
            self.log.debug(
                "Failed to read devel group membership for user '%s': %s",
                self.user, stderr
            )
        if not ret and stdout:
            return True
        return False

    def remote_repository_exists(self, repo):
        """Verifies whether repository for specific package is available.
        :param str repo: repository name
        :return: True if repository exists, else False
        :rtype: bool
        """
        if not repo:
            raise Exception("Repository name is empty")
        git_url = self._get_namespace_anongiturl(repo)
        cmd = [  # lists artifacts (only HEAD) of the remote repository
            'git',
            'ls-remote',
            git_url,
            'HEAD'
        ]
        ret, __, stderr = self._run_command(cmd, return_stdout=True, return_stderr=True)
        if ret and stderr:
            self.log.debug(
                "Failed to reach remote repository '%s': %s:",
                repo, stderr
            )
        return not bool(ret)

    def mock_config(self, *args, **kwargs):
        """
        overrided mock_config just for patching its output
        """
        config = super(Commands, self).mock_config(*args, **kwargs)
        # !WORKAROUND!: for rhel6.X releases patch mockbuild config.
        # There is an incompatibility with systemd-nspawn, therefore
        # specific option is added to config. JIRA: COMPOSE-2223
        if re.search(r'\brhel-6.\d+', self.branch_merge):
            config += '\nconfig_opts[\'use_nspawn\'] = False'
        return config


class BrewService(object):
    """
    brew buildsystem related methods
    """

    def __init__(self, session, logger):
        # xmlrpc session
        self.session = session
        self.log = logger

    def _get_collection_prefix(self, s):
        """
        s can be target or dist-git branch
          rhscl-X.Y-$prefix-rhel-{6,7}
          rhscl-X.Y-$prefix-rhel-{6,7}-candidate
        """
        regex = re.compile(r"^\w+-\d+\.\d+-(?P<prefix>.+?)-\w+-\w+(?:-candidate)?$")
        m = regex.match(s)
        try:
            return m.group('prefix')
        except AttributeError:
            raise RuntimeError("'%s' is not a rhscl dist-git branch or target." % s)

    def latest_collection_build_info(self, package, target, tag):
        """
        software collections are special, we have to figure out latest build a bit different

        we take dist-git branch and extract prefix from it:

          rhscl-X.Y-$prefix-rhel-{6,7}

        then we query brew like: brew latest-build $tag $prefix-$package
        """
        # try to fetch extra data from brew. If data is missing,
        # fallback to standard scl prefix detection
        extra_info = self.get_extra_info(target)
        if extra_info and extra_info.get("scl"):
            prefix = extra_info["scl"]
            self.log.debug('Using scl prefix gathered from Brew extra info: %s', prefix)
        else:
            prefix = self._get_collection_prefix(target)
        scl_package = "%s-%s" % (prefix, package)
        build = self.session.getLatestBuilds(tag, package=scl_package)
        try:
            build_info = build[0]
        except IndexError:
            raise RuntimeError("No build found for package '%s' tagged by '%s'." %
                               (scl_package, tag))
        return build_info

    def get_tag(self, target):
        """
        get tag from provided target:

        brew list-targets | grep $target | awk '{ print $3 }'
        """
        self.log.debug("Get tag for target '%s'.", target)
        target_info = self.session.getBuildTarget(target)
        if target_info:
            return target_info['dest_tag_name']
        else:
            raise RuntimeError("Can't find target '%s'." % target)

    def get_extra_info(self, target):
        """
        get extra info fields from provided target.
        Do not raise any exception, return empty result instead.
        """
        self.log.debug("Get tag for target '%s'.", target)
        target_info = self.session.getTag(target)
        if target_info:
            search = target_info["id"]
            # at this point, search is a build tag's ID
            conf = self.session.getBuildConfig(search)
            extra = conf.get("extra", {})
        else:
            extra = {}

        # copy just explicit fields
        extra_info = {
            'rpm_macro_dist': extra.get("rpm.macro.dist", "").lstrip(".") or None,
            'rhpkg_dist': extra.get("rhpkg_dist"),
            'scl': extra.get("rhpkg_scl_prefix"),
        }

        return extra_info

    def latest_build_info(self, package, target, tag=None):
        """
        get latest build info of package from brew
        """
        if tag is None:
            tag = self.get_tag(target)
            self.log.debug("Using brew tag '%s'.", tag)

        build = self.session.getLatestBuilds(tag, package=package)

        try:
            build_info = build[0]
        except IndexError:
            # could be rhscl branch, lets try it
            build_info = self.latest_collection_build_info(package, target, tag)
        return build_info

    def latest_build(self, package, target, tag=None):
        """
        return latest build for provided package

        you have to specify target, which is "${dist-git-branch}-candidate" -- it is
        used for SCL prefix extraction if no build is found for package tagged by tag
        """
        build_info = self.latest_build_info(package, target, tag)
        try:
            nvr = build_info['nvr']
        except KeyError as ex:
            raise RuntimeError("Invalid response from build system: %s" % ex)
        self.log.info("Latest build of package '%s', in tag '%s' is '%s'", package, tag, nvr)
        return nvr

    def get_build_info(self, build):
        """
        return information for build from brew
        build can be specified by ID or NVR
        """
        return self.session.getBuild(build)

    def get_build_commit(self, build):
        """
        return commit hash of build
        build can be specified by ID or NVR
        """
        build_info = self.get_build_info(build)
        try:
            task_id = build_info['task_id']
        except (KeyError, TypeError) as ex:
            raise RuntimeError("Invalid response from build system: %s" % ex)
        task_info = self.session.getTaskInfo(task_id, True)
        return self.get_commit_hash(task_info)

    def get_latest_build_commit(self, package, target, tag=None):
        """
        return commit has of last build
        """
        build_info = self.latest_build_info(package, target, tag)
        try:
            task_id = build_info['task_id']
        except (KeyError, TypeError) as ex:
            raise RuntimeError("Invalid response from build system: %s" % ex)

        task_info = self.session.getTaskInfo(task_id, True)
        return self.get_commit_hash(task_info)

    def get_commit_hash(self, task_info):
        """
        return commit hash from task information
        """
        try:
            request = task_info['request'][0]
        except (KeyError, IndexError) as ex:
            raise RuntimeError("Invalid response from build system: %s" % ex)
        # commit hash is after '#'
        commit_hash = request.split('#')[1]
        return commit_hash
