#!/usr/bin/python
# rhpkg - a script to interact with the Red Hat Packaging system
#
# Copyright (C) 2011 Red Hat Inc.
# Author(s): Jesse Keating <jkeating@redhat.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

import logging
import os
import ssl
import sys

import requests
import six.moves.configparser as ConfigParser

import pyrpkg
import pyrpkg.utils
import rhpkg.cli
from pyrpkg import rpkgError
from rhpkg import SSL_VERIFY_FAILED_HINT


def show_hint_for_failure_ssl_verify(e, log):
    is_ssl_verify_err = (
        isinstance(e, ssl.SSLError) and e.reason == 'CERTIFICATE_VERIFY_FAILED' or
        # Raised from Koji
        isinstance(e, requests.exceptions.SSLError) and 'CERTIFICATE_VERIFY_FAILED' in str(e) or
        # Raised from rpmdiff-remote
        isinstance(e, rpkgError) and "Peer's Certificate issuer is not recognized." in str(e)
    )
    if is_ssl_verify_err:
        ssl_verify_hint = 'Certificate verification failed.{0}'.format(SSL_VERIFY_FAILED_HINT)
        log.warn(ssl_verify_hint)


def main():
    # Setup an argparser and parse the known commands to get the config file
    program_name = os.path.basename(sys.argv[0])

    # Modified ArgumentParser provides parameter "allow_abbrev=False" (which affects processing
    # of commandline arguments with common prefix). Generaly it is available since python3.6.
    # This enables "allow_abbrev" for older python versions.
    parser = pyrpkg.cli.ArgumentParser(add_help=False)
    parser.add_argument('-C', '--config', help='Specify a config file to use',
                        default='/etc/rpkg/%s.conf' % program_name)

    (args, other) = parser.parse_known_args()

    # Make sure we have a sane config file
    if not os.path.exists(args.config) and not other[-1] in ['--help', '-h', 'help']:
        sys.stderr.write('Invalid config file %s\n' % args.config)
        sys.exit(1)

    # Setup a configuration object and read config file data
    config = ConfigParser.SafeConfigParser()
    config.read(args.config)

    client = rhpkg.cli.rhpkgClient(config)
    client.do_imports(site='rhpkg')
    client.parse_cmdline()

    # This is due to a difference argparse behavior to Python 2 version.
    # In Python 3, argparse will proceed to here without reporting
    # "too few arguments". Instead, client.args does not have attribute
    # command.
    if not hasattr(client.args, 'command'):
        client.parser.print_help()
        sys.exit(1)

    if not client.args.path:
        try:
            client.args.path = pyrpkg.utils.getcwd()
        except Exception:
            print('Could not get current path, have you deleted it?')
            sys.exit(1)

    # setup the logger -- This logger will take things of INFO or DEBUG and
    # log it to stdout.  Anything above that (WARN, ERROR, CRITICAL) will go
    # to stderr.  Normal operation will show anything INFO and above.
    # Quiet hides INFO, while Verbose exposes DEBUG.  In all cases WARN or
    # higher are exposed (via stderr).
    log = pyrpkg.log
    client.setupLogging(log)

    if client.args.v:
        log.setLevel(logging.DEBUG)
    elif client.args.q:
        log.setLevel(logging.WARNING)
    else:
        log.setLevel(logging.INFO)

    # Run the necessary command
    try:
        sys.exit(client.args.command())
    except KeyboardInterrupt:
        pass
    except Exception as e:
        show_hint_for_failure_ssl_verify(e, log)

        if getattr(client.args, 'debug', False):
            raise
        log.error('Could not execute %s: %s' % (client.args.command.__name__, e))
        sys.exit(1)


if __name__ == "__main__":
    main()
