# -*- coding: utf-8 -*-

"""
Functions for interaction between dist-git and bugzilla.

Inspired by project rpt.

Copyright (C) 2014 Red Hat Inc.

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
the full text of the license.
"""

import bugzilla

from .service import DryRunService, LoggingService

COMMENT_FIXEDIN_TEMPLATE = """\
FIXED IN
--------
%s
"""

COMMENT_VCS_LINKS_TEMPLATE = """\
PATCH(ES)
---------
%s
"""


class BugzillaService(LoggingService, DryRunService):
    """
    High-level methods for interaction with bugzilla
    """

    def __init__(self, url, username=None, password=None, *args, **kwargs):
        """
        url [string]               -- bugzilla URL
        username/password [string] -- credentials for bugzilla (not required, cookie works too)
        """
        LoggingService.__init__(self, *args, **kwargs)
        DryRunService.__init__(self, *args, **kwargs)
        self.bz = bugzilla.Bugzilla(url=url, user=username, password=password)

    def modified(self, bug_id):
        """
        Change state of provided bugzilla to modified
        """
        return self.batch_update(bug_id, to_modified=True)

    def _format_fixed_field(self, builds):
        """
        Put list of builds into required common format
        """
        return " ".join(builds)

    def _format_comment(self, fixed_in_nvrs, vcs_urls):
        """
        Format data for informational comment
        """
        global COMMENT_FIXEDIN_TEMPLATE, COMMENT_VCS_LINKS_TEMPLATE
        parts = []
        if fixed_in_nvrs:
            parts.append(COMMENT_FIXEDIN_TEMPLATE % "\n".join(fixed_in_nvrs))
        if vcs_urls:
            parts.append(COMMENT_VCS_LINKS_TEMPLATE % "\n".join(vcs_urls))
        return "\n".join(parts)

    def set_fixed_in(self, bug_id, builds):
        """
        Set fixed in version field

        bug_id [string/int]      -- bug to update
        builds [list of strings] -- list of nvrs to put into Fixed-In field
        """
        return self.batch_update(bug_id, fixed_in_nvrs=builds, restrict_nvrs=True)

    def add_info_comment(self, bug_id, builds, urls):
        """
        Compose a comment with Fixed In field data and links to VCS

        bug_id [string/int]      -- bug to update
        builds [list of strings] -- list of nvrs to put into Fixed-In field
        urls [list of strings]   -- list of link(s) to commit(s)
        """
        return self.batch_update(bug_id, fixed_in_nvrs=builds, commit_urls=urls, restrict_nvrs=True)

    def _format_query(self, query):
        """
        Don't print ugly dict, format it!
        """
        response = ""
        if 'comment' in query:
            response = '\nCOMMENT:\n' + '=' * 80 + '\n' + \
                       query['comment']['body'] + '=' * 80 + '\n'
        if 'cf_fixed_in' in query:
            response += 'FIXED IN = %s\n' % query['cf_fixed_in']
        if 'status' in query:
            response += 'STATUS = %s\n' % query['status']
        return response

    def batch_update(
        self,
        bug_id,
        fixed_in_nvrs=None,
        to_modified=False,
        commit_urls=None,
        restrict_nvrs=False
    ):
        """
        update multiple fields within single call

        bug_id [string/int]             -- bug to update
        fixed_in_nvrs [list of strings] -- list of nvrs to put into Fixed-In field
        to_modified [bool]              -- change state to modified?
        commit_urls [list of strings]   -- list of link(s) to commit(s)
        restrict_nvrs [bool]            -- restrict modification of already set Fixed-In field
        """
        # Fetch bug
        bug = self.bz.getbug(bug_id)

        # Create a query
        query = {}

        if fixed_in_nvrs:
            formatted_nvrs = self._format_fixed_field(fixed_in_nvrs)
            if restrict_nvrs:
                # do not modify 'Fixed In Version' field if it was already set
                if bug.cf_fixed_in:
                    self.log.error("Bug #%s has already filled 'Fixed In Version' field: '%s'.",
                                   bug_id, bug.cf_fixed_in)
                else:
                    query['cf_fixed_in'] = formatted_nvrs
            else:
                # show only warning and modify 'Fixed In Version' field
                if bug.cf_fixed_in:
                    self.log.warning("Bug #%s has already filled 'Fixed In Version' field: '%s'. "
                                     "Will be modified to '%s'",
                                     bug_id, bug.cf_fixed_in, formatted_nvrs)
                query['cf_fixed_in'] = formatted_nvrs

        if commit_urls:
            body = self._format_comment(fixed_in_nvrs, commit_urls)
            query['comment'] = {
                'body': body,
                'is_private': True,
            }

        if to_modified:
            acc_states = ('NEW', 'ASSIGNED', 'POST')
            cur_state = bug.status
            if cur_state in acc_states:
                query['status'] = 'MODIFIED'
            else:
                self.log.error("Bug #%s should be in one of these states: %s, but it's in '%s'",
                               bug_id, acc_states, cur_state)

        # Batch update the bugzilla iff the length of query is non-zero
        if len(query):
            if not self.dry_log("I would update bug '%s', with these settings:\n%s",
                                bug_id, self._format_query(query)):
                # returns "{'bugs': [{'alias': [], 'changes': {'status':..."
                response = self.bz.update_bugs([bug_id], query)
                self.log.info("Bug was updated:\n%s", self._format_query(query))
                return response

    def get_status(self, bug_id):
        """
        call that reads bug info

        :param string/int bug_id: bug to read
        :return: None if bug record doesn't exists or error appears, else status
        :rtype: string
        """
        if not self.dry_log("I would get status of bug '#%s'", bug_id):
            try:
                bug = self.bz.getbug(bug_id)
            except Exception as e:
                self.log.error("Bugzilla bug '#%s' can not be read: '%s'.", bug_id, e)
                raise
            return getattr(bug, "status", None)
