# cli.py - a cli client class module for rhpkg
#
# Copyright (C) 2011 Red Hat Inc.
# Author(s): Jesse Keating <jkeating@redhat.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

import argparse
import io
import json
import os
import random
import re
import string
import subprocess
import sys
import time

import six
import six.moves.configparser as ConfigParser
from requests.exceptions import HTTPError
from six import iteritems

import koji
import koji_cli.lib
import pycurl
from pyrpkg import rpkgError
from pyrpkg.cli import cliClient
from pyrpkg.utils import make_koji_watch_tasks_handler

from .httpsession import HTTPSession


class rhpkgClient(cliClient):
    def __init__(self, config, name=None):
        self.DEFAULT_CLI_NAME = 'rhpkg'
        super(rhpkgClient, self).__init__(config, name)
        self.setup_rh_subparsers()

    def setup_rh_subparsers(self):
        self.register_parser()
        self.register_remote()
        self.register_brew_path()
        self.register_maven_build()
        self.register_maven_chain()
        self.register_wrapper_rpm()
        self.register_tests()
        self.register_publican_build()
        self.register_covscan()
        self.register_errata()
        self.register_rpmdiff()
        self.register_bugzilla()
        self.register_ci_submit_result()

    # Target registry goes here
    # First add a global option
    def register_parser(self):
        """This is to add extra global options to the main parser"""

        self.parser.add_argument('--released', action='store_true',
                                 default='DISCOVER',
                                 help='Specify that the product being built '
                                      'for has been released, which could '
                                      'determine the value of the dist tag.')

        self.parser.add_argument('--unreleased', action='store_false',
                                 default='DISCOVER', dest='released',
                                 help='Specify that the product being built '
                                      'for has not been released yet, which '
                                      'could determine the value of the dist '
                                      'tag.')

    def register_remote(self):
        """Add command remote and options"""
        remote_parser = self.subparsers.add_parser(
            'remote',
            help='Add repository from remote dist-git services.')

        remote_subparsers = remote_parser.add_subparsers(
            dest='remote_action',
            description='Currently, rhpkg allows you to add remote (that is the git-remote) and'
                        ' operate conveniently on top of the remote. WARNING: all remote '
                        'commands are experimental. Behavior and CLI options may change in the'
                        ' future.')
        remote_add_parser = remote_subparsers.add_parser(
            'add',
            description='Add a new repository from remote dist-git service. Currently, Fedora and'
                        ' CentOS are supported.')

        remote_add_subparsers = remote_add_parser.add_subparsers(dest='remote_name')

        remote_backends = [(name.split('_')[0], value) for name, value
                           in self.config.items('{0}.remotes'.format(self.name))
                           if name.endswith('_cli')]
        for remote, backend_name in remote_backends:
            remote_add_fedora_parser = remote_add_subparsers.add_parser(
                remote,
                description='Add a remote to {0} dist-git. This requires to have {1} '
                            'installed.'.format(remote, backend_name))
            remote_add_fedora_parser.add_argument('repo_name', nargs='?', help='repository name')

        remote_parser.set_defaults(command=self.remote)

    def register_brew_path(self):
        """Register the brew-path target"""

        brew_path_parser = self.subparsers.add_parser(
            'brew-path',
            help='Print the path to the brew builds for the current package '
                 'n-v-r',
            description='This will print a path to where the build output can '
                        'be found in the brew filesystem. It uses the current '
                        'n-v-r listed in the spec file.')
        brew_path_parser.set_defaults(command=self.brew_path)

    def register_maven_build(self):
        """Register the maven-build target"""

        self.maven_build_parser = self.subparsers.add_parser(
            'maven-build',
            parents=[self.build_parser_common],
            help='Build a package in MEAD',
            description='Build a package in MEAD')
        excl_group = self.maven_build_parser.add_mutually_exclusive_group(
            required=True)
        excl_group.add_argument(
            '--sources',
            metavar='URL',
            help='URL to the source: svn+http://...?...#...')
        src_group = self.maven_build_parser.add_argument_group(
            title='sources',
            description='Valid options when using --sources')
        src_group.add_argument(
            '--patches',
            metavar='URL',
            help='URL to where the patches are used for the current SCM '
                 'branch')
        src_group.add_argument('--goal', action='append',
                               dest='goals', metavar='GOAL',
                               help='Additional goal to run before "deploy"')
        src_group.add_argument('--profile', action='append',
                               dest='profiles', metavar='PROFILE',
                               help='Enable a Maven profile')
        src_group.add_argument('--property', action='append',
                               dest='properties', metavar='NAME=VALUE',
                               help='Set a Java system property')
        src_group.add_argument('--env', action='append',
                               dest='envs', metavar='NAME=VALUE',
                               help='Set an environment variable')
        src_group.add_argument('--package', action='append',
                               dest='packages', metavar='PACKAGE',
                               help='Additional package to install in the '
                                    'buildroot')
        src_group.add_argument('--jvm-option', action='append',
                               dest='jvm_options', metavar='OPTION',
                               help='Pass an option to the JVM')
        src_group.add_argument('--maven-option', action='append',
                               dest='maven_options', metavar='OPTION',
                               help='Pass an option to Maven')
        src_group.add_argument('--specfile',
                               metavar='URL',
                               help='URL to where the .spec.tmpl file is, '
                                    'use . for the current SCM branch')
        excl_group.add_argument('--ini', action='append',
                                dest='inis', metavar='CONFIG',
                                help='Provide build parameters via a .ini '
                                     'file')
        ini_group = self.maven_build_parser.add_argument_group(
            title='ini',
            description='Valid options when using --ini')
        ini_group.add_argument('--section',
                               help='Get build parameters from this section '
                                    'of the .ini file')
        common_group = self.maven_build_parser.add_argument_group(
            title='common',
            description='Common options')
        common_group.add_argument('--debug', action='store_true',
                                  help='Run Maven in debug mode')
        common_group.add_argument('--skip-tag', action='store_true',
                                  help='Do not attempt to tag build')
        common_group.add_argument('--scratch', action='store_true',
                                  help='Perform a scratch build')
        self.maven_build_parser.set_defaults(command=self.maven_build)

    def register_maven_chain(self):
        """Register the maven-chain target"""

        self.maven_chain_parser = self.subparsers.add_parser(
            'maven-chain',
            parents=[self.build_parser_common],
            help='Build a group of Maven packages in dependency order',
            description='Build a group of Maven packages in dependency order')
        self.maven_chain_parser.add_argument(
            '--ini',
            action='append',
            required=True,
            dest='inis',
            metavar='CONFIG',
            help='.ini file containing build parameters')
        self.maven_chain_parser.add_argument(
            '--debug',
            action='store_true',
            help='Run Maven in debug mode')
        self.maven_chain_parser.add_argument(
            '--skip-tag',
            action='store_true',
            help='Do not attempt to tag build')
        self.maven_chain_parser.add_argument(
            '--scratch',
            action='store_true',
            help='Perform a scratch build')
        self.maven_chain_parser.set_defaults(command=self.maven_chain)

    def register_wrapper_rpm(self):
        """Register the wrapper-rpm target"""

        wrapper_rpm_parser = self.subparsers.add_parser(
            'wrapper-rpm',
            parents=[self.build_parser_common],
            help='Build wrapper rpms for an existing Brew build',
            description='Build wrapper rpms for an existing Brew build')
        wrapper_rpm_parser.add_argument(
            '--build',
            required=True,
            help='NVR or ID of the build to create wrapper rpms for')
        wrapper_rpm_parser.add_argument(
            '--specfile',
            metavar='URL', default='.',
            help='URL to where the .spec.tmpl file is, defaults to the current'
                 ' SCM branch')
        wrapper_rpm_parser.add_argument(
            '--no-create-build',
            action='store_false',
            dest='create_build',
            help="Don't create a new build to contain wrapper rpms")
        wrapper_rpm_parser.add_argument(
            '--skip-tag',
            action='store_true',
            help='Do not attempt to tag build')
        wrapper_rpm_parser.add_argument(
            '--scratch',
            action='store_true',
            help='Perform a scratch build')
        wrapper_rpm_parser.set_defaults(command=self.wrapper_rpm)

    def register_tests(self):
        """Register the tests target"""

        tests_parser = self.subparsers.add_parser(
            'tests',
            help='Clone and checkout a tests module',
            description='This command will clone the named tests module from '
                        'the configured repository base URL. By default it '
                        'will also checkout the master branch for your '
                        'working copy.')
        # provide a convenient way to get to a specific branch
        tests_parser.add_argument('--branch', '-b',
                                  help='Check out a specific branch')
        # allow to clone without needing a account on the scm server
        tests_parser.add_argument('--anonymous', '-a',
                                  action='store_true',
                                  help='Check out a module anonymously')
        # store the module to be cloned
        tests_parser.add_argument(
            'module',
            nargs='?',
            default='DISCOVER',
            help='Name of the tests module to clone. If no module name is '
                 'given, it will be guessed by the current directory name.')
        tests_parser.set_defaults(command=self.tests)

    # Overloaded register functions here
    def register_build(self):
        # Do all the work from the super class
        super(rhpkgClient, self).register_build()
        # Now add our additional option
        self.subparsers.choices['build'].add_argument(
            '--fast',
            action='store_true',
            default=False,
            help='Perform build as a FasTrack update. This will alter the '
                 'build target used.')

    def register_publican_build(self):
        """Register the publican-build target"""

        pubbuild_parser = self.subparsers.add_parser(
            'publican-build',
            help='Commit and build a publican package',
            description='This command will do the leg work of updating a '
                        'publican package in brew.')
        # set the language
        pubbuild_parser.add_argument(
            '--lang',
            required=True,
            default=None,
            help='Build for language (use "none" for splash pages)')
        # provide a changelog message
        pubbuild_parser.add_argument(
            '--message', '-m',  # TODO: pull from rpm spec changelog
            default='Docs updated by Dude McPants',
            help='Specify a commit message')

        # scratch build?
        pubbuild_parser.add_argument('--scratch',
                                     help='Only do a scratch build',
                                     action='store_true')
        # watch output?
        pubbuild_parser.add_argument('--nowait',
                                     help="Don't wait on build",
                                     action='store_true')
        # set the publican config file
        pubbuild_parser.add_argument('--publican-config',
                                     help='Publican config file',
                                     default='publican.cfg')
        # other publican arguments (needed for --desktop)
        pubbuild_parser.add_argument('--publican-args',
                                     help='Supply extra arguments to publican',
                                     default=None)
        # provide a convenient way to get to a specific branch
        pubbuild_parser.add_argument('--branch', '-b',
                                     help='Use a specific branch',
                                     default=None)
        # override target?
        pubbuild_parser.add_argument('--target',
                                     help='Override the default target',
                                     default=None)
        pubbuild_parser.set_defaults(command=self.publican_build)

    def register_covscan(self):
        covscan_parser = self.subparsers.add_parser(
            'covscan',
            help='Perform static analysis scans',
            description='You may initiate scans in covscan via this command. '
                        'If you provide base, the scan will be differential. '
                        'WARNING: This command is experimental. Behaviour and'
                        ' CLI options may change in the future.',
                        # '$ rhpkg covscan --base "foo.src.rpm" '
                        # '--covscan-args "--security --concurrency"'
        )
        covscan_parser.add_argument(
            '--mock-profile', '-r',
            default=None,
            help='Specify mock profile (otherwise it will be figured out from '
                 'branch name)')
        covscan_parser.add_argument(
            '--covscan-args',
            default=None,
            metavar="'--some --args'",
            help="Custom arguments for covscan (defaults to "
                 "'--profile=errata')")
        covscan_parser.add_argument(
            '--base', '-b',
            help="Initiate diff scan and use this srpm or brew build as a base"
                 " (latest build from brew will be used if srpm or brew build "
                 "is not provided)",
            metavar="foo.src.rpm or NVR",
            nargs='?',
            default="")
        covscan_parser.add_argument(
            '--check-patches',
            action='store_true',
            help="Scan downstream patches")
        covscan_parser.set_defaults(command=self.covscan)

    def register_rpmdiff(self):
        rpmdiff_parser = self.subparsers.add_parser(
            'rpmdiff',
            help='Perform scratch Rpmdiff scans',
            description='You may initiate scratch Rpmdiff scans prior to '
                        'production Rpmdiff in Errata Tool. If you provide a '
                        'base, the scan will be differential. Analysis is '
                        'performed on scratch builds, so you have to provide '
                        'an ID of the task. WARNING: This command is '
                        'experimental. Behaviour and CLI options may change '
                        'in the future.')
        rpmdiff_parser.add_argument('scratch_id',
                                    type=int,
                                    metavar="scratch build ID",
                                    help="ID of scratch build")
        rpmdiff_parser.add_argument(
            '--base', '-b',
            metavar="foo-1-1.el7",
            nargs="?",
            default="",
            help="Initiate diff scan and use this build as a base. Latest build from Brew will be "
                 "used if you don't provide it.")
        rpmdiff_parser.set_defaults(command=self.rpmdiff)

    def register_errata(self):
        self.errata_parser = self.subparsers.add_parser(
            'errata',
            help="Manage your erratum",
            # arguments with common suffix are not "ambiguous" now (--bugzilla_username vs. --bug)
            allow_abbrev=False,
            description="If you do not specify the erratum and product version, the tool will try "
                        "to figure it out by searching for unreleased erratum for the current "
                        "dist-git repo and branch. You SHOULD try --dry-run first before telling "
                        "the tool to change something. For adding bugzilla bugs, some IDs require "
                        "login to bugzilla service. You do not have to specify bugzilla username "
                        "and password; python-bugzilla caches session information in the "
                        "'~/.cache/python-bugzilla/' folder. You can use '$ bugzilla login' to "
                        "populate the file with cookies. WARNING: This command is experimental. "
                        "Behaviour and CLI options may change in the future.")
        self.errata_parser.add_argument('--erratum',
                                        help="Specify ID of an erratum",
                                        default=None,
                                        metavar="ERRATUM-ID")
        self.errata_parser.add_argument('--dry-run',
                                        help="Don't change anything, just say what would you do",
                                        action="store_true",
                                        default=None)
        self.errata_parser.add_argument('--bugzilla-username',
                                        help="Bugzilla username (for additional "
                                             "verification of errata queries)",
                                        default=None)
        self.errata_parser.add_argument('--bugzilla-password',
                                        help="Bugzilla password",
                                        default=None)

        self.errata_subparser = self.errata_parser.add_subparsers(dest='errataDest')

        self.errata_state = self.errata_subparser.add_parser('new-state',
                                                             help="Change erratum state")
        self.errata_state.add_argument('state', choices=['qe', 'new-files'])
        self.errata_state.add_argument('--message', '-m', help='New status message', default=None)

        self.errata_add_bugs = self.errata_subparser.add_parser('add-bugs',
                                                                help="Add bugs to erratum")
        self.errata_add_bugs_excl_group = self.errata_add_bugs.add_mutually_exclusive_group()
        self.errata_add_bugs_excl_group.add_argument('--last-commit',
                                                     help="Use bugs in latest commit (Default)",
                                                     action="store_true")
        self.errata_add_bugs_excl_group.add_argument('--since-last-build',
                                                     help="Use bugs since latest build",
                                                     action="store_true")
        self.errata_add_bugs_excl_group.add_argument('--since-commit',
                                                     help="Use bugs since specified commit",
                                                     default=None,
                                                     metavar="COMMIT")
        self.errata_add_bugs_excl_group.add_argument('--since-build',
                                                     help="Use bugs since specified build",
                                                     default=None,
                                                     metavar="BUILD")
        self.errata_add_bugs_excl_group.add_argument('--bug',
                                                     help="Use specified bug IDs",
                                                     metavar="BUG#",
                                                     nargs="+",
                                                     dest="bugs")

        self.errata_remove_bugs = self.errata_subparser.add_parser(
            'remove-bugs',
            help="Provide a list of bugs to remove from erratum")
        self.errata_remove_bugs.add_argument('bugs', nargs="+", metavar="BUG#", default=None)

        self.errata_add_builds = self.errata_subparser.add_parser(
            'add-builds',
            help="Add build to erratum (you may specify nvr)")
        self.errata_add_builds.add_argument('builds', metavar="N-V-R", nargs="*", default=None)
        self.errata_add_builds.add_argument(
            '--product-version',
            default=None,
            dest='product_versions',
            metavar='VERSION',
            action='append',
            help="Specify product version ID (for example 'RHEL-6'). "
                 "You can specify this option multiple times in order to "
                 "add builds to several product versions at once.")
        self.errata_add_builds.add_argument('--nowait-rpmdiff',
                                            help="Don't wait for rpmdiff job. The erratum will "
                                                 "not be moved to QE state automatically.",
                                            action="store_true")

        # Following two options are usefull only for testing against
        # non-Kerberos enabled Errata Tool instances
        # (e.g. errata.app.qa.eng.nay.redhat.com)
        self.errata_parser.add_argument('--username', help=argparse.SUPPRESS)
        self.errata_parser.add_argument('--password', help=argparse.SUPPRESS)
        self.errata_parser.set_defaults(command=self.errata)

    def register_bugzilla(self):
        self.bugzilla_parser = self.subparsers.add_parser(
            'bugzilla',
            help="Update bugzilla with required information",
            description="You do not have to specify username and password; python-bugzilla caches "
                        "session information in the '~/.cache/python-bugzilla/' folder. You can "
                        "use '$ bugzilla login' to populate the file with cookies. If you do not "
                        "specify the Bugzilla ID, the one from the last commit will be used. "
                        "If you do not provide a list of NVRs for the 'Fixed In Version' field, "
                        "the latest build will be used (it is a good idea to build it first "
                        "and then update Bugzilla). WARNING: This command is experimental. "
                        "Behaviour and CLI options might change in the future.")
        self.bugzilla_parser.add_argument('--username',
                                          help="Bugzilla username",
                                          default=None)
        self.bugzilla_parser.add_argument('--password',
                                          help="Bugzilla password",
                                          default=None)
        self.bugzilla_parser.add_argument('--dry-run',
                                          help="Don't change anything, just "
                                               "say what would you do",
                                          action="store_true",
                                          default=None)
        action_group = self.bugzilla_parser.add_argument_group(title='actions')
        action_group.add_argument('--fixed-in',
                                  help="Update 'Fixed In Version' field "
                                       "(multiple NVRs can be specified; if no"
                                       " NVR is specified, the latest build "
                                       "is used)",
                                  default=None,
                                  nargs="*",
                                  metavar="N-V-R")
        action_group.add_argument('--modified',
                                  help="Change state of Bugzilla to modified",
                                  default=None,
                                  action="store_true")
        action_group.add_argument('--comment',
                                  help="Add a comment to Bugzilla (a list of "
                                       "patches, build NVR, and so on.)",
                                  default=None,
                                  action="store_true")
        bug_selector_group = \
            self.bugzilla_parser.add_argument_group(title='selectors')
        bug_selector_excl_group = \
            bug_selector_group.add_mutually_exclusive_group()
        bug_selector_excl_group.add_argument(
            '--last-commit',
            help="Use bugs in latest commit (Default)",
            action="store_true")
        bug_selector_excl_group.add_argument(
            '--since-last-build',
            help="Use bugs since latest build",
            action="store_true")
        bug_selector_excl_group.add_argument(
            '--since-commit',
            help="Use bugs since specified commit",
            default=None,
            metavar="COMMIT")
        bug_selector_excl_group.add_argument(
            '--since-build',
            help="Use bugs since specified build",
            default=None,
            metavar="BUILD")
        bug_selector_excl_group.add_argument(
            '--bug',
            help="Use specified bug IDs",
            metavar="BUG#",
            nargs="+",
            dest="bugs")
        self.bugzilla_parser.set_defaults(command=self.bugzilla)

    def register_sources(self):
        """Extend sources command

        With this extension, there are three forms of sources as below

            rhpkg sources
            rhpkg sources fedora
            rhpkg sources fedora f24

        Where, the 3rd form is to allow download source files from a
        particular remote branch.
        """
        super(rhpkgClient, self).register_sources()

        valid_remotes = [name.replace('_cli', '') for name, value in
                         self.config.items('{0}.remotes'.format(self.name))
                         if name.endswith('_cli')]
        valid_remotes = ' and '.join(valid_remotes)

        sources_subparser = self.subparsers.choices['sources']
        sources_subparser.description = 'Download source files. Exploded sources is also ' \
                                        'supported, that allows to download source files ' \
                                        'from outside repositories listed in source-repos file. ' \
                                        'For more information about Exploded Sources, please ' \
                                        'refer to https://mojo.redhat.com/docs/DOC-1059964.'
        sources_subparser.add_argument(
            'remote_name',
            nargs='?',
            help='remote name. Valid names are {0}.'.format(valid_remotes))
        sources_subparser.add_argument(
            'branch',
            nargs='?',
            help='remote and branch must be specified at same time. With these'
                 ' two arguments, rhpkg can know from which to get source '
                 'files. Also, both of these argument are optional.')

    def register_mockbuild(self):
        super(rhpkgClient, self).register_mockbuild()

        mockbuild_parser = self.subparsers.choices['mockbuild']

        # Here is a magic trick. The reason why dest is named remote_name is to
        # reuse `sources` method to download sources properly, from either
        # remote or internal dist-git.
        # `sources` method recognizes remote name from an option also named
        # remote_name, so to give same name here could be much convenient and
        # straightforward.
        mockbuild_parser.add_argument(
            '--sources-from',
            dest='remote_name',
            choices=self.get_support_remotes(),
            help='from where to download sources, that is useful when multiple'
                 ' remotes are added.')

    def register_ci_submit_result(self):
        help_msg = 'Submit a test result for Manual Gating workflow'
        description = '''Submits a test result for Manual Gating workflow.

To submit a manual test result, you need to run the command with all required parameters:

    rhpkg ci-submit-result --testcase manual.sst.sst_cockpit.manual-integration-test \
--result passed --url https://jira/browse/PROJECT-1234 --nvr cockpit-1.0.0-1.el8

For more information about Manual Gating workflow see the documentation here:

    https://docs.engineering.redhat.com/display/RHELPLAN/Manual+Gating+workflow

'''

        ci_submit_result_parser = self.subparsers.add_parser(
            'ci-submit-result',
            formatter_class=argparse.RawDescriptionHelpFormatter,
            help=help_msg,
            description=description)
        ci_submit_result_parser.add_argument(
            '--nvr', '-n', required=True,
            help='Build NVR for which the result is for'),
        ci_submit_result_parser.add_argument(
            '--result', '-r', required=True,
            help='The result of the testing (passed or failed)',
            choices=['passed', 'failed'])
        ci_submit_result_parser.add_argument(
            '--testcase', '-t', required=True,
            help=(
                'The testcase name as defined according to the documentation. '
                'For example: manual.sst_cockpit.cockpit.'
                'manual-integration-test'
            ))
        ci_submit_result_parser.add_argument(
            '--url', '-u', required=True,
            help=(
                'Link to a Bugzilla or Jira issue with the attached'
                ' test results.'
            ))
        ci_submit_result_parser.set_defaults(command=self.ci_submit_result)

    def get_support_remotes(self):
        return [name.split('_')[0] for name, value
                in self.config.items('{0}.remotes'.format(self.name))
                if name.endswith('_cli')]

    # Target functions go here
    def brew_path(self):
        print(self.cmd.brew_path)

    def maven_build(self):
        target_override = False
        # Override the target if we were supplied one
        if self.args.target:
            self.cmd._target = self.args.target
            target_override = True

        if self.args.inis and \
                (self.args.patches or self.args.goals or self.args.profiles or
                 self.args.specfile or self.args.envs or self.args.packages or
                 self.args.jvm_options or self.args.maven_options or
                 self.args.properties):
            self.maven_build_parser.error(
                'parameters may not be passed on the command-line when '
                'using --ini')

        # Initiate the build task
        try:
            task_id = self.cmd.maven_build(
                self.args.sources,
                patches=self.args.patches,
                goals=self.args.goals,
                profiles=self.args.profiles,
                properties=self.args.properties,
                envs=self.args.envs,
                packages=self.args.packages,
                jvm_options=self.args.jvm_options,
                maven_options=self.args.maven_options,
                inis=self.args.inis,
                section=self.args.section,
                specfile=self.args.specfile,
                debug=self.args.debug,
                background=self.args.background,
                skip_tag=self.args.skip_tag,
                scratch=self.args.scratch,
                target_override=target_override)
        except Exception as e:
            self.log.error('Could not initiate maven-build: %s', e)
            sys.exit(1)

        # Now that we have the task ID we need to deal with it.
        if self.args.nowait:
            # Log out of the koji session
            self.cmd.kojisession.logout()
            return
        # pass info off to our koji task watcher
        try:
            self.cmd.kojisession.logout()
            return koji_cli.lib.watch_tasks(
                self.cmd.kojisession,
                [task_id],
                ki_handler=make_koji_watch_tasks_handler(self.cmd.build_client)
            )
        except Exception:
            # We could get an auth error if credentials have expired
            # use exc_info here to get what kind of error this is
            self.log.error('Could not watch maven-build: %s', sys.exc_info()[0])
            sys.exit(1)

    def maven_chain(self):
        target_override = False
        # Override the target if we were supplied one
        if self.args.target:
            self.cmd._target = self.args.target
            target_override = True

        task_id = self.cmd.maven_chain(self.args.inis,
                                       debug=self.args.debug,
                                       background=self.args.background,
                                       skip_tag=self.args.skip_tag,
                                       scratch=self.args.scratch,
                                       target_override=target_override)
        self.cmd.kojisession.logout()
        if not self.args.nowait:
            return koji_cli.lib.watch_tasks(
                self.cmd.kojisession,
                [task_id],
                ki_handler=make_koji_watch_tasks_handler(self.cmd.build_client)
            )

    def wrapper_rpm(self):
        # Override the target if we were supplied one
        if self.args.target:
            self.cmd._target = self.args.target

        # Initiate the build task
        try:
            task_id = self.cmd.wrapper_rpm(self.args.build,
                                           self.args.specfile,
                                           create_build=self.args.create_build,
                                           background=self.args.background,
                                           skip_tag=self.args.skip_tag,
                                           scratch=self.args.scratch)
        except Exception as e:
            self.log.error('Could not initiate wrapper-rpm: %s', e)
            sys.exit(1)

        # Now that we have the task ID we need to deal with it.
        if self.args.nowait:
            # Log out of the koji session
            self.cmd.kojisession.logout()
            return
        # pass info off to our koji task watcher
        try:
            self.cmd.kojisession.logout()
            return koji_cli.lib.watch_tasks(
                self.cmd.kojisession,
                [task_id],
                ki_handler=make_koji_watch_tasks_handler(self.cmd.build_client)
            )
        except Exception:
            # We could get an auth error if credentials have expired
            # use exc_info here to get what kind of error this is
            self.log.error('Could not watch wrapper-rpm: %s', sys.exc_info()[0])
            sys.exit(1)

    def tests(self):
        self.log.warning('tests command is deprecated. Please use clone command '
                         'instead, e.g. rhpkg clone tests/{0}.'.format(
                             self.args.module.replace('tests/', '')))

        if self.args.module == 'DISCOVER':
            self.log.debug('Using cwd as module name')
            self.args.module = os.path.basename(os.getcwd())
        self.cmd.tests(self.args.module, branch=self.args.branch,
                       anon=self.args.anonymous)

    # Overloaded target functions go here
    def load_cmd(self):
        """Overload to add more attributes"""

        # Call the super
        super(rhpkgClient, self).load_cmd()
        if self.args.released != 'DISCOVER':
            self.cmd.released = self.args.released
        self.cmd.cgit_url_template = self.config.get(self.name + '.dist-git',
                                                     'commit_url',
                                                     raw=True)
        try:
            self.cmd.pp_url = self.config.get(self.name + '.ProductPages',
                                              'url',
                                              raw=True)
        except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
            self.cmd.pp_url = None

    def build(self, sets=None):
        if hasattr(self.args, 'fast') and self.args.fast:
            # Check if we have target and fast options
            if self.args.target:
                raise Exception('Cannot mix --target and --fast')
            self.args.target = self.cmd.target.replace('-candidate', '-fast-candidate')

        if self.args.target is None:
            self.cmd.check_target_with_release_plan()

        # Call the super build
        return super(rhpkgClient, self).build(sets)

    def scratch_build(self):
        # Set the fast to false and then do the normal stuff
        self.args.fast = False
        return super(rhpkgClient, self).scratch_build()

    def container_build(self, flatpak=False):
        if self.args.target is None:
            self.cmd.check_target_with_release_plan(container_build=True)

        return super(rhpkgClient, self).container_build(flatpak)

    def publican_build(self):
        # check publican config exists
        if not os.path.exists(self.args.publican_config):
            raise Exception('No publican config (%s).' % (
                self.args.publican_config))

        # determine branch from publican config if possible
        if self.args.branch is None:
            conf = ConfigParser.SafeConfigParser()
            publican_config_header = "[publican]\n"
            publican_config_body = open(self.args.publican_config).read()
            publican_config_section = publican_config_header + publican_config_body
            if not isinstance(publican_config_section, six.text_type):
                # in case of Python 2 byte strings have to be converted to unicode
                publican_config_section = publican_config_section.decode('utf-8')
            conf.readfp(io.StringIO(publican_config_section))
            if conf.has_option("publican", "git_branch"):
                self.args.branch = conf.get("publican", "git_branch").strip()
            elif conf.has_option("publican", "cvs_branch"):
                self.args.branch = conf.get(
                    "publican", "cvs_branch").lower().strip()
            del conf

        if self.args.scratch:
            # check there's a target specified
            if self.args.target is None:
                if self.args.branch is not None:
                    self.args.target = self.args.branch + "-candidate"
                else:
                    raise Exception("Must specify a target for scratch builds")

        if not self.args.scratch:
            # check there's a commit message of some sort
            if self.args.message is None:
                raise Exception("Must specify a commit message")

            if self.args.branch is None:
                raise Exception("No branch specified (use --branch or "
                                "git_branch: in publican.cfg")

        # invoke publican to build an srpm
        srpm_path = self.cmd.publican_package(
            self.args.lang, self.args.publican_config, self.args.publican_args)

        if self.args.scratch:
            # for a scratch build, upload the srpm
            callback = None
            if not self.args.q:
                callback = koji_cli.lib._progress_callback
            uniquepath = 'cli-build/%r.%s' % (
                time.time(),
                ''.join([random.choice(string.ascii_letters)
                         for i in range(8)]))
            self.cmd.koji_upload(srpm_path, uniquepath, callback=callback)
            if not self.args.q:
                print('')
            url = '%s/%s' % (uniquepath, os.path.basename(srpm_path))
        else:
            # for a non-scratch build, update dist-git
            self.cmd.publican_git_commit(srpm_path, self.args.branch,
                                         self.args.message)
            url = None

        if self.args.target is not None:
            self.cmd._target = self.args.target

        # do the build
        task_id = self.cmd.build(False, self.args.scratch, False, url, None,
                                 None, None)

        # don't need to be logged in any longer
        self.cmd.kojisession.logout()

        # watch the build
        if not self.args.nowait:
            return koji_cli.lib.watch_tasks(
                self.cmd.kojisession,
                [task_id],
                ki_handler=make_koji_watch_tasks_handler(self.cmd.build_client)
            )

    def covscan(self):
        mock_profile = self.args.mock_profile
        covscan_args = self.args.covscan_args
        check_patches = self.args.check_patches
        # base_srpm_path = 'foo.src.rpm'  # use this src.rpm as base_srpm_path
        # no argument provided to base_srpm_path, use latest build
        # base_srpm_path = None
        # default value, base_srpm_path argument was not specified
        # base_srpm_path = ''
        base_srpm_path = self.args.base
        executable = self.config.get(self.name + '.covscan', 'executable')

        if check_patches and \
                (base_srpm_path is None or len(base_srpm_path) > 0):
            raise RuntimeError("You have specified '--check-patches' and "
                               "'--base_srpm_path', these options are mutually"
                               " exclusive.")

        # we allow '--covscan-args=' to disable using the default args
        # (currently --profile=errata)
        if covscan_args is None:
            covscan_args_list = None
        else:
            # we are using re.split because covscan fails to parse such
            # options:
            # e.g. ' --all'
            covscan_args_list = re.split(r'\s+', covscan_args.strip())
        # FIXME: there should be an easier way to create a srpm
        self.args.hash = 'md5'
        self.srpm()
        srpm_path = self.cmd.srpmname

        latest = None
        # base_srpm_path is specified, but no argument provided -- use latest
        # brew build
        if base_srpm_path is None:
            latest = self.cmd.brew_service.latest_build(
                target=self.cmd.target,
                package=self.cmd.module_name)
        # NVR of brew build is provided with --base option
        elif not base_srpm_path.endswith('src.rpm') and base_srpm_path != '':
            build_info = self.cmd.kojisession.getBuild(self.args.base)
            if not build_info:
                raise RuntimeError('Build \"%s\" does not exist in Brew' %
                                   self.args.base)
            latest = self.args.base
            base_srpm_path = None
        return self.cmd.covscan(srpm_path,
                                mock_profile=mock_profile,
                                covscan_args=covscan_args_list,
                                base_srpm_path=base_srpm_path,
                                latest=latest,
                                check_patches=check_patches,
                                executable=executable)

    def rpmdiff(self):
        base = self.args.base
        scratch_id = self.args.scratch_id
        hub_url = self.config.get(self.name + '.test-rpmdiff', 'hub')

        task = self.cmd.kojisession.getTaskInfo(scratch_id)
        state = task['state']
        koji_in_process_states = (koji.TASK_STATES['ASSIGNED'],
                                  koji.TASK_STATES['FREE'],
                                  koji.TASK_STATES['OPEN'])
        if state in koji_in_process_states:
            self.log.info("Waiting for build [%s]", scratch_id)
            build_ret_val = koji_cli.lib.watch_tasks(
                self.cmd.kojisession,
                [int(scratch_id)],
                ki_handler=make_koji_watch_tasks_handler(self.cmd.build_client)
            )
            if build_ret_val != 0:
                raise RuntimeError("Scratch build [%s] failed - %d" % (scratch_id, build_ret_val))

        if state == koji.TASK_STATES['FAILED']:
            self.log.warning('Scratch build {0} is failed.'.format(scratch_id))

        # base is specified, but no argument provided -- use latest brew build
        if base is None:
            base = self.cmd.brew_service.latest_build(target=self.cmd.target,
                                                      package=self.cmd.module_name)
        elif base != '':
            build = self.cmd.brew_service.get_build_info(base)
            if build is None:
                raise RuntimeError('Build {0} does not exist in Brew.'.format(base))
        else:
            # --base is not specified, so an analysis run will be scheduled.
            base = None

        return self.cmd.rpmdiff(hub_url, scratch_id, base=base)

    def errata(self):
        """
        Parse command line arguments for errata command and do the job
        """
        et_conf = dict(self.config.items(self.name + '.errata', raw=True))
        umb_config = dict(self.config.items(self.name + '.UMB', raw=True))
        umb_config['brokers'] = umb_config['brokers'].split()

        # options
        erratum_id = self.args.erratum
        bugzilla_username = self.args.bugzilla_username
        bugzilla_password = self.args.bugzilla_password
        dry_run = self.args.dry_run
        username = self.args.username
        password = self.args.password
        debug = self.args.debug
        verbose = self.args.v

        # set default values
        move_to_qe = False
        move_to_newfiles = False
        message = None
        product_versions = None
        builds = None
        bugs_to_add = None
        bugs_to_remove = None
        nowait_rpmdiff = False

        # according to sub-option set appropriate arguments
        if self.args.errataDest == 'new-state':
            move_to_qe = self.args.state == "qe"
            move_to_newfiles = self.args.state == "new-files"
            message = self.args.message
        elif self.args.errataDest == 'add-builds':
            product_versions = self.args.product_versions
            builds = self.args.builds
            # build is not set.. get latest build from brew
            if builds == []:
                builds = [self.cmd.brew_service.latest_build(target=self.cmd.target,
                                                             package=self.cmd.module_name)]
            nowait_rpmdiff = self.args.nowait_rpmdiff
        elif self.args.errataDest == 'remove-bugs':
            bugs_to_remove = self.args.bugs
        elif self.args.errataDest == "add-bugs":
            bugs_to_add = self.get_bugs_according_option(self.args.since_last_build,
                                                         self.args.since_build,
                                                         self.args.since_commit,
                                                         self.args.bugs)

        # errata command uses bugzilla cli to verify bugs
        try:
            bugzilla_url = self.config.get(self.name + '.bugzilla', 'url')
        except Exception:
            raise RuntimeError("Invalid configuration for bugzilla, missing URL.")

        self.cmd.errata(
            et_conf,
            umb_config,
            erratum=erratum_id,
            product_versions=product_versions,
            bugs_to_add=bugs_to_add,
            bugs_to_remove=bugs_to_remove,
            builds=builds,
            move_to_newfiles=move_to_newfiles,
            move_to_qe=move_to_qe,
            dry_run=dry_run,
            debug=debug,
            verbose=verbose,
            username=username,
            password=password,
            message=message,
            nowait_rpmdiff=nowait_rpmdiff,
            bugzilla_url=bugzilla_url,
            bugzilla_username=bugzilla_username,
            bugzilla_password=bugzilla_password,
        )

    def check_if_builds_include_bug_fixes(self, builds, commits):
        commits_hash = set([str(commit) for commit in commits])
        build_commits = dict((nvr, self.cmd.brew_service.get_build_commit(nvr))
                             for nvr in builds)
        for nvr, commit_hash in iteritems(build_commits):
            proc = subprocess.Popen(['git', 'rev-list', commit_hash],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    cwd=self.cmd.path,
                                    universal_newlines=True)
            stdout, stderr = proc.communicate()
            commits_outside_build = commits_hash - set(stdout.strip().split('\n'))
            if commits_outside_build:
                raise rpkgError('Build {0} does not contain fix {1}'.format(
                    nvr, ', '.join(commits_outside_build)))

    def bugzilla(self):
        """
        Parse arguments for bugzilla command, figure out missing and fire
        update
        """
        self.cmd.check_repo(is_dirty=False)

        username = self.args.username
        password = self.args.password
        fixed_in = self.args.fixed_in
        modified = self.args.modified
        comment = self.args.comment
        bug_ids = self.args.bugs
        dry_run = self.args.dry_run
        since_last_build = self.args.since_last_build
        since_build = self.args.since_build
        since_commit = self.args.since_commit
        last_commit = self.args.last_commit

        if(fixed_in is None and modified is None and comment is None and
                bug_ids is None and since_last_build is False and
                last_commit is False and since_build is None and
                since_commit is None):
            self.bugzilla_parser.error("too few arguments")

        try:
            url = self.config.get(self.name + '.bugzilla', 'url')
        except Exception:
            raise RuntimeError("Invalid configuration for bugzilla, missing URL.")

        if fixed_in == []:
            fixed_in = [self.cmd.brew_service.latest_build(target=self.cmd.target,
                                                           package=self.cmd.module_name)]

        bug_ids = self.get_bugs_according_option(since_last_build,
                                                 since_build, since_commit,
                                                 bug_ids)
        form_commit_url = self.cmd.distgit_service.form_commit_url
        for bug_id in bug_ids:
            commit_urls = []
            if comment or fixed_in:
                commits = self.cmd.distgit_service.list_commits_for_bug(bug_id)
                if not commits:
                    raise RuntimeError("No commits found for bug '%s'." % bug_id)
                if fixed_in:
                    self.check_if_builds_include_bug_fixes(fixed_in, commits)
                for commit in commits:
                    commit_url = form_commit_url(self.cmd.ns_repo_name, commit)
                    commit_urls.append(commit_url)

            if not fixed_in and not modified and not comment:
                raise RuntimeError("No update options specified. Nothing to do.")

            self.cmd.bugzilla(
                url,
                bug_id,
                username=username,
                password=password,
                fixed_in_nvrs=fixed_in,
                to_modified=modified,
                commit_urls=commit_urls,
                dry_run=dry_run,
            )

    def get_bugs_according_option(self, since_last_build, since_build,
                                  since_commit, bug_ids):
        # Get list of bugs according to bugzilla or errata options
        if since_last_build:
            last_build_commit = self.cmd.brew_service.get_latest_build_commit(
                target=self.cmd.target, package=self.cmd.module_name)
            bug_ids = self.cmd.distgit_service.get_bugs_since_commit(last_build_commit)
        elif since_build:
            build_commit = self.cmd.brew_service.get_build_commit(since_build)
            bug_ids = self.cmd.distgit_service.get_bugs_since_commit(build_commit)
        elif since_commit:
            bug_ids = self.cmd.distgit_service.get_bugs_since_commit(since_commit)
        elif not bug_ids:
            bug_ids = self.cmd.distgit_service.get_last_bugs()

        return bug_ids

    def clone(self):
        try:
            super(rhpkgClient, self).clone()
        except rpkgError:
            # Perform some postchecks to provide more info to user:
            # 1) Does repository which has to be cloned even exist?
            repo = self.args.repo[0]  # name of repository; mandatory param. for clone
            repository_exists = self.cmd.remote_repository_exists(repo)
            if not repository_exists:
                self.log.error(
                    "\nRepository '%s' is not reachable, most likely it does not exist.",
                    repo
                )
            # 2) Is user a member of 'devel' group?
            devel_member = self.cmd.user_is_devel_member()
            if not devel_member:
                self.log.error(
                    "\nCurrent user '%s' is not member of 'devel' LDAP group "
                    "(or maybe LDAP is down).",
                    self.cmd.user
                )
            raise

    def remote(self):
        """Handle command remote"""
        section_name = '{0}.remotes'.format(self.name)
        remote_backends = dict(self.config.items(section_name))

        self.cmd.remote(self.args.remote_name, self.args.repo_name, remote_backends)

    def sources(self):
        section_name = '{0}.remotes'.format(self.name)
        remote_backends = dict(self.config.items(section_name))

        self.cmd.sources(outdir=getattr(self.args, 'outdir', None),
                         remote=getattr(self.args, 'remote_name', None),
                         branch=getattr(self.args, 'branch', None),
                         remote_backends=remote_backends)

    def mockbuild(self):
        """Extended mockbuild to support multiple remotes

        sources files must be downloaded before starting a mock build. This
        extended version of `mockbuild` calls `sources` method at first,
        because it is good at handling multiple remotes instead of pyrpkg's
        `sources` method.

        Based on current implementation of pyrpkg, it's difficult to override
        pyrpkg's `mockbuild` (I don't want to duplicate most of the code from
        `pyrpkg.Command.mockbuild`). That causes `rhpkg.Command.sources` will
        be called twice here. But, it's safe and okay without any side-effect.

        Note: if the logic of checking whether a sources file exists and has
        same hash is changed, please consider the necessarity to modify or
        rewrite this method.
        """
        # For mockbuild, it's no need to specify output directory and the
        # remote branch.
        self.args.outdir = None
        self.args.branch = None
        self.sources()

        super(rhpkgClient, self).mockbuild()

    def ci_submit_result(self):
        """Handle submitting a test result for Manual Gating Workflow"""
        nvr = self.args.nvr
        testcase = self.args.testcase
        result = self.args.result
        url = self.args.url

        # Validate testcase name
        testcase_regex = r'manual.sst_[\w-]+.[\w-]+.[\w-]+$'
        if not re.match(testcase_regex, testcase):
            raise RuntimeError(
                "The testcase name '{0}' is invalid.".format(testcase))

        # get email from current user
        email = '{0}@redhat.com'.format(self.cmd.user)

        # get brew task ID of the given build
        try:
            buildinfo = self.cmd.kojisession.getBuild(nvr)
        except HTTPError as e:
            raise RuntimeError(
                "Error getting build '{0}' info: {1}".format(nvr, e))

        if not buildinfo:
            raise RuntimeError("Unknown build '{0}'".format(nvr))

        # try to get MBS build id, to know if this is a module build
        try:
            module_info = buildinfo['extra']['typeinfo']['module']
            mbs_build_id = module_info['module_build_service_id']
        except KeyError:
            mbs_build_id = None

        # prepare details for module builds
        if mbs_build_id:
            nsvc = '{name}:{stream}:{version}:{context}'.format(**module_info)

            submit_result_option = 'submit_result_module_url'

            specific_parameters = [
                {'name': 'build_id', 'value': str(mbs_build_id)},
                {'name': 'nsvc', 'value': nsvc}
            ]

        else:
            submit_result_option = 'submit_result_url'

            specific_parameters = [
                {'name': 'task_id', 'value': str(buildinfo['task_id'])},
                {'name': 'nvr', 'value': nvr}
            ]

        # get submit result url for rpm or module builds
        try:
            submit_result_url = self.config.get(
                self.name + '.ci',
                submit_result_option)
        except ConfigParser.NoSectionError as e:
            raise RuntimeError(
                'Missing required configuration in rhpkg.conf: {0}'.format(e))

        params = {
            'parameter': [
                # general parameters
                {'name': 'testcase', 'value': testcase},
                {'name': 'result', 'value': result},
                {'name': 'run_url', 'value': url},
                {'name': 'email', 'value': email},
                {'name': 'MSG_PROVIDER', 'value': 'Red Hat UMB'}
            ] + specific_parameters
        }

        # submit result, use GSSAPI
        session = HTTPSession(debug=self.args.debug, log=self.log)

        try:
            retcode, response = session.request(
                submit_result_url, 'post', data="json={0}".format(
                    json.dumps(params)))
        except pycurl.error as e:
            raise RuntimeError('Could not report result: {0}'.format(e))

        if retcode != 201:
            error = (
                'Submitting result failed, rerun rhpkg with debug option '
                'and contact OSCI team.\n\nemail: '
                'osci-list@redhat.com, irc: #osci\n\n'
            )
            raise RuntimeError(
                error + 'return code: {0}\n\nresponse: {1}'.format(
                    retcode, response if response else 'N/A'))

        self.log.info(
            'Result successfully enqueued and will be processed soon.')
