# -*- encoding: utf-8 -*-

"""
module for covscan service

Copyright (C) 2014 Red Hat Inc.
Author(s): Tomas Tomecek <ttomecek@redhat.com>

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
the full text of the license.
"""

import os
import re
import subprocess

from .service import LoggingService


def form_mock_profile(branch):
    """
    Try to create mock config from provided dist-git branch
    """
    try:
        stripped_branch = re.findall(r"rhel-\d+", branch)[0]
    except IndexError:
        raise RuntimeError("Error: can't figure out mock profile from dist-git branch.")
    mock_template = "%s-x86_64"
    mock_profile = mock_template % stripped_branch
    return mock_profile


class CovscanService(LoggingService):
    """
    interface for direct interaction with Covscan tool
    """

    def __init__(self, executable, *args, **kwargs):
        super(CovscanService, self).__init__(*args, **kwargs)
        self.executable = executable
        with open(os.devnull, 'w') as devnull:
            try:
                subprocess.check_call(['which', self.executable], stdout=devnull, stderr=devnull)
            except subprocess.CalledProcessError:
                raise RuntimeError(
                    "Can't find executable '%s' See "
                    "https://gitlab.cee.redhat.com/covscan/covscan/wikis/covscan" %
                    self.executable)

    def _run(self, *args):
        """
        run covscan with provided args, don't process stdout so user knows what's going on
        """
        cmd = [self.executable] + list(args)
        self.log.debug("Running: '%s'", " ".join(cmd))
        out = subprocess.call(cmd)
        return out

    def _run_and_return_output(self, *args):
        """
        run covscan with provided args, return everything from stdout
        """
        cmd = [self.executable] + list(args)
        self.log.debug("Running: %s", " ".join(cmd))
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        out = p.communicate()[0]
        return out.strip()

    def list_mock_configs(self):
        """
        return list of enabled mock profiles on covscan service
        """
        cfgs = self._run_and_return_output("list-mock-configs").split('\n')
        enabled_cfgs = []
        for cfg_line in cfgs:
            cfg, enabled = cfg_line.split()
            if enabled == 'True':
                enabled_cfgs.append(cfg)
        return enabled_cfgs

    def _production_args(self, mock_profile, covscan_args=None):
        """
        production configuration of covscan
        """
        cs_args = [
            "--config=%s" % mock_profile,
            "--comment=submitted by rhpkg",
        ]
        if covscan_args is None:
            cs_args += [
                "--profile=errata",
            ]
        else:
            cs_args += covscan_args
        return cs_args

    def scan(self, mock_profile, srpm_path, covscan_args=None):
        """
        scan package with production configuration
        """
        cs_args = self._production_args(mock_profile, covscan_args)
        final_args = ['mock-build'] + cs_args + [srpm_path]
        self._run(*final_args)

    def diff_scan(self, mock_profile, srpm_path, covscan_args=None):
        """
        scan patches
        """
        cs_args = self._production_args(mock_profile, covscan_args)
        final_args = ['diff-build'] + cs_args + [srpm_path]
        self._run(*final_args)

    def version_diff_scan(self, mock_profile, srpm_path, base_srpm_path=None, base_brew_build=None,
                          covscan_args=None):
        """
        do a differential scan
        """
        cs_args = self._production_args(mock_profile, covscan_args)
        final_args = ['version-diff-build'] + cs_args + ['--srpm=' + srpm_path, ]
        if base_srpm_path:
            final_args += ['--base-srpm=' + base_srpm_path]
        elif base_brew_build:
            final_args += ['--base-brew-build=' + base_brew_build]
        else:
            raise RuntimeError("Specify base srpm or brew build please.")
        self._run(*final_args)


class CovscanManager(object):
    """
    functions to interact with covscan
    """

    def __init__(self, executable, dist_git_branch, mock_profile=None, log=None, covscan_args=None):
        self.branch = dist_git_branch
        self.covscan_args = covscan_args
        self.mock_profile = mock_profile or form_mock_profile(self.branch)
        self.cs = CovscanService(executable, log=log)

    def analyze(self, srpm_path):
        """
        scan provided srpm in covscan
        """
        self.cs.scan(self.mock_profile, srpm_path, covscan_args=self.covscan_args)

    def diff_analyze(self, srpm_path):
        """
        do a diffbuild task -- analyze patches
        """
        self.cs.diff_scan(self.mock_profile, srpm_path, covscan_args=self.covscan_args)

    def version_diff_analyze(self, srpm_path, base_srpm_path=None, base_brew_build=None):
        """
        submit differential scan to covscan

        target is 'srpm_path'
        base_brew_build is nvr of production build in brew (not scratch build)
        either base_srpm_path or base_brew_build has to be specified, not both
        """
        self.cs.version_diff_scan(self.mock_profile, srpm_path, base_srpm_path=base_srpm_path,
                                  base_brew_build=base_brew_build, covscan_args=self.covscan_args)
