# -*- coding: utf-8 -*-

"""
Dist-git related functions.

Copyright (C) 2014 Red Hat Inc.
Author(s): Tomas Tomecek <ttomecek@redhat.com>

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
the full text of the license.
"""


import re

from git import GitCommandError

from .service import LoggingService


class DistgitService(LoggingService):
    """
    functions related to dist-git
    """

    def __init__(self, repo, cgit_url_template, *args, **kwargs):
        super(DistgitService, self).__init__(*args, **kwargs)
        self.repo = repo
        self.cgit_url_template = cgit_url_template

    def _get_bugs_from_str(self, s):
        """Try to find bugs in provided string"""
        BZRE = re.compile(r'(?:(?:bug|bz|rhbz)\s*#?|#)\s*(\d+)', re.IGNORECASE)
        regex = re.compile(r'^\s*-?\s*(?:Resolves?:|Related?:|Reverts?:)(.*)',
                           re.IGNORECASE | re.MULTILINE)
        bzs = []
        for match in regex.finditer(s):
            for bz_id in BZRE.finditer(match.group(1)):
                bzs.extend(bz_id.groups())
        return bzs

    def get_last_bugs(self):
        """
        Find last bug_ids -- return list of bug_id from latest commit that
        matches "Resolved/Related: <bug_id>"
        """
        found = None
        for commit in self.repo.iter_commits():
            found = self._get_bugs_from_str(commit.message)
            if found:
                break
        if not found:
            raise RuntimeError("No bug found in commit messages")
        self.log.info("Last bugs are: %s", ", ".join(found))
        return found

    def get_bugs_since_commit(self, commit_hash):
        bug_list = []
        if not self.commit_hash_exists(commit_hash):
            self.log.warning("Commit %s does not exist in repository" % commit_hash)
        else:
            for commit in self.repo.iter_commits():
                bugs = self._get_bugs_from_str(commit.message)
                bug_list.extend(bugs)
                commit_short = self.repo.git.rev_parse(commit.hexsha, short=1)
                if commit_short == self.repo.git.rev_parse(commit_hash, short=1):
                    break
        if not bug_list:
            raise RuntimeError("No bug found in commit messages")
        self.log.info("Last bugs are: %s", ", ".join(bug_list))
        return bug_list

    def commit_hash_exists(self, commit_hash):
        """
        Check if commit exists in repo
        """
        try:
            self.repo.git.rev_parse(commit_hash, short=1)
        except GitCommandError:
            return False
        return True

    def list_commits_for_bug(self, bug_id):
        """
        Return all commits related to provided bug
        """
        matching = []
        for commit in self.repo.iter_commits():
            if bug_id in self._get_bugs_from_str(commit.message):
                matching.append(commit)
        return matching

    def form_commit_url(self, module, commit):
        """
        Provide clickable links for commits
        """
        if not self.cgit_url_template:
            raise RuntimeError("You have to specify cgit URL.")
        try:
            return self.cgit_url_template % {
                "module": module,
                "commit_hash": commit.hexsha,
            }
        except KeyError as exc:
            raise RuntimeError(
                'commit_url template in config file contains unknown placeholder %s\n'
                'Only "module" and "commit_hash" placeholders are allowed.'
                % exc)
