# -*- coding: utf-8 -*-

"""
Functions to manage your erratas.

I would like to thank Simon Baird <sbaird@redhat.com> for helping me out to
figure out some stuff.

Copyright (C) 2014 Red Hat Inc.
Author(s): Tomas Tomecek <ttomecek@redhat.com>

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
the full text of the license.
"""

from json import dumps, loads

import kobo.xmlrpc
from kobo.xmlrpc import SafeCookieTransport
from six.moves import xmlrpc_client
from six.moves.urllib.parse import urljoin

from .bugzilla_service import BugzillaService
from .httpsession import HTTPSession
from .messagebus import ETRPMDiffMessageWatcher
from .service import DryRunService, LoggingService


class ET(LoggingService):
    def __init__(self, conf, debug=False, verbose=False, username=None,
                 password=None, *args, **kwargs):
        """
        conf has to contain key 'url' that is configured in rhpkg.conf pointing
        to an Errata Tool instance.

        See :class:`httpsession.HTTPSession` for username and password.
        """
        super(ET, self).__init__(debug=debug, verbose=verbose, *args, **kwargs)
        # "https://errata.devel.redhat.com/"
        # "https://errata-devel.app.eng.bos.redhat.com/"
        try:
            self.base = conf['url']
        except KeyError:
            raise RuntimeError(
                'Missing URL to Errata Tool in configuration %s.' % conf)
        self.rest_url = urljoin(self.base, "api/v1/")
        self.web_url = self.base
        self._xmlrpc = None
        # curl http sessions, one for JSON and one not
        self.www_session = HTTPSession(debug=debug, log=self.log,
                                       username=username, password=password)
        self.rest_session = HTTPSession(debug=debug, log=self.log, json=True,
                                        username=username, password=password)

    @property
    def xmlrpc(self):
        if self._xmlrpc is None:
            if not kobo.xmlrpc.USE_KERBEROS:
                raise RuntimeError('Request to Errata Service requires '
                                   'kerberos package but it was not found.')
            self._xmlrpc = xmlrpc_client.ServerProxy(
                urljoin(self.base, "errata/xmlrpc.cgi"),
                transport=SafeCookieTransport())
            # test if connection is okay
            self.xmlrpc.ping()
        return self._xmlrpc

    def _build_rest_url(self, path):
        return urljoin(self.rest_url, path)

    def _handle_response_if_request_fails(self, status_code, response_data):
        """Ensure to raise error if request to REST API fails

        Error handling is documented[1]. If the status code is in the
        documented scope, error must be handled and raise RuntimeError with
        proper error message.

        [1] https://errata.devel.redhat.com/developer-guide/api-http-api.html#api-error-handling
        """
        if status_code not in (401, 403, 404, 422, 400, 500):
            return

        error = []
        if 'errors' in response_data:
            self.log.error('Server Responded with errors: %s', response_data['errors'])
            error.append(str(response_data['errors']))
        if 'error' in response_data:
            self.log.error("An error has occured: %s", response_data['error'])
            error.append(str(response_data['error']))
        raise RuntimeError('Invalid request: %s' % ''.join(error))

    def _do_rest_request(self, method, path, data=None):
        """
        initiate a request on REST API on `path` using `method` and sending
        `data`
        """
        if data:
            data = dumps(data)

        url = self._build_rest_url(path)

        self.log.debug('[%s] %s', method, url)
        self.log.debug('[data] %s', data)

        status_code, response_text = self.rest_session.request(method=method,
                                                               url=url,
                                                               data=data)

        self.log.debug('[response] %s', status_code)
        self.log.debug('Raw Response: %s', response_text)

        if status_code == 401 or status_code == 403:
            # FIXME: as of writing this code, Errata Tool API responses a piece
            # of HTML instead of JSON data when 401 happens. Bug 1339896 has
            # been reported about this issue. After it gets fixed (I hope so),
            # it's no need to handle 401 here specifically.
            response_data = {
                'error': 'Authorization of request to Errata Tool is required.'
            }
        else:
            response_data = loads(response_text) if response_text else None

        self._handle_response_if_request_fails(status_code, response_data)

        return status_code, response_data

    def get_advisory(self, e_id):
        """Fetch advisory/erratum -- same thing

        ET REST API erratum/{id} returns a big JSON data containing requested
        erratum information. Following is an example of return value with
        partial data.

        {
            ...
            'errata': {
                'rhsa': {
                    'errata_id': 20050,
                    'status': 'NEW_FILES',
                    'group_id': 452,
                }
            }
            ...
        }

        By talking with Errata Tool developer, currently, data['errata']['rhsa'] is the requested
        erratum.
        """
        path = 'erratum/%s' % (e_id)
        status_code, response_data = self._do_rest_request('get', path)
        return list(response_data['errata'].values())[0]

    def get_release(self, release_id):
        path = 'releases/%s' % (release_id)
        status_code, response_data = self._do_rest_request('get', path)
        return response_data

    def add_bug(self, e_id, bug_id):
        path = 'erratum/%s/add_bug' % (e_id)
        return self._do_rest_request('post', path, {
            'bug': bug_id
        })

    def remove_bug(self, e_id, bug_id):
        path = 'erratum/%s/remove_bug' % (e_id)
        return self._do_rest_request('post', path, {
            'bug': bug_id
        })

    def add_builds(self, e_id, product_id, nvrs):
        path = 'erratum/%s/add_builds' % (e_id)
        return self._do_rest_request('post', path, [{
            'product_version': product_id,
            'build': nvr,
            'file_types': None
        } for nvr in nvrs])

    def change_state(self, e_id, state, message):
        path = 'erratum/%s/change_state' % (e_id)
        return self._do_rest_request('post', path, {
            'new_state': str(state),
            'comment': message
        })

    def sync_bug(self, bug_id):
        path = 'bug/refresh'
        return self._do_rest_request('post', path, [bug_id])

    def find_erratum(self, package_name):
        return self.xmlrpc.get_advisory_list({
            # package is not reliable, b/c if erratum does not contain any
            # build
            # ET doesn't know that the erratum is associated with the package
            # "package": package_name,

            # we want to match: "|package |bug fix and..." so we won't catch
            # packages with names as prefixes, e.g. mutt and mutter
            "synopsis": package_name + " ",
            # lets not include REL_PREP, since it's about to be released and we
            # don't expect new builds
            'statuses': ["QE", "NEW_FILES"],
            })


class ChoiceException(Exception):
    """
    We have detected there are multiple choices and we don't want to pick one.
    Let user decide.
    """


class ErrataManager(LoggingService, DryRunService):
    """
    Encapsulate all ET related methods in one class.
    This is the high level API

    Most methods should work without specifying 'erratum' and/or 'product'.
    """

    def __init__(self, et_conf, umb_config, package=None, erratum=None,
                 product_versions=None, debug=False, verbose=False,
                 username=None, password=None,
                 bugzilla_url=None, bugzilla_username=None, bugzilla_password=None,
                 *args, **kwargs):
        """
        :param dict et_conf: configuration of errata tool containing base address
        :param dict umb_config: configuration of message bus where rpmdiff sends updates
        :param str package: package name to work with (used for searching for erratum)
        :param number erratum: ID of erratum
        :param str product_version: ID of product version
        :param bool dry_run: if true, don't change anything, just print what would you do

        See :class:`httpsession.HTTPSession` for username and password.
        """
        LoggingService.__init__(self, *args, **kwargs)
        DryRunService.__init__(self, *args, **kwargs)

        self.package = package

        # options
        self.erratum = erratum
        self.product_versions = product_versions

        # lets cache json response with info about erratum
        self.erratum_json = None

        if package is None and erratum is None:
            raise RuntimeError("Please, specify either package or erratum.")

        self.umb_config = umb_config
        self.et_conf = et_conf
        self.et = ET(et_conf, log=self.log, debug=debug, verbose=verbose,
                     username=username, password=password, bugzilla_url=bugzilla_url,
                     bugzilla_username=bugzilla_username, bugzilla_password=bugzilla_password,
                     bugzilla_dry_run=self.dry_run)

        # bugzilla cli for verification bugs
        self.bz = BugzillaService(bugzilla_url, username=bugzilla_username,
                                  password=bugzilla_password, log=self.log, dry_run=self.dry_run)

    def _erratum_setdefault(self):
        """Try to get erratum from ET and cache it

        If erratum is given from command line, it'll be used to query an
        erratum data by calling ET XMLRPC API ``get_advisory``.

        If erratum is omitted, module name, that is named package in this
        class, got from current dist-git repo is used to find its associated
        erratum data. However, this does not work all the time, because an
        advisory is not associated with a package before a build is added.

        Whatever the way above to get erratum data, raw JSON data is saved
        without changes for any further use.
        """
        if self.erratum_json is not None:
            return
        if self.erratum is None:
            self.erratum_json = self._find_erratum()
            self.erratum = self.erratum_json['errata_id']
        else:
            self.erratum_json = self.et.get_advisory(self.erratum)

    def _product_setdefault(self):
        """ cache product """
        if self.product_versions is None:
            self.product_versions = [self._find_product()]

    def _find_erratum(self):
        """
        find erratum from current dist-git repo, provide package if it is not
        related to current dist-git repo in case of multiple erratas, raise
        Exception

        return whole errata JSON
        """
        self.log.debug("Looking for erratum...")

        def do_output_if_multiple_erratas(errata_list):
            """ create output for Exc if multiple erratas found """
            result = "We have found mulitple matching erratas, please pick " \
                     "correct one:\n"
            for e in errata_list:
                result += "  * %(errata_id)s - [%(product)s] %(synopsis)s " \
                          "(%(status)s, created at %(created_at)s)\n" % e
            return result

        erratas = self.et.find_erratum(self.package)
        if len(erratas) < 1:
            self.log.error("Error: no erratas found")
            raise RuntimeError("No erratas found. If your errata doesn't "
                               "exist, create it please.")
        elif len(erratas) == 1:
            try:
                self.log.info("Using erratum: '%s' (%s)",
                              erratas[0]['errata_id'], erratas[0]['synopsis'])
            except KeyError:
                self.log.error("Error: Invalid response from ET, missing key "
                               "'errata_id' or 'synopsis'")
                raise RuntimeError("Invalid response from ET.")
            return erratas[0]
        else:
            self.log.error("Error: multiple erratas found")
            raise ChoiceException(do_output_if_multiple_erratas(erratas))

    def _find_product(self):
        """
        find product version for erratum; in case of multiple products, raise
        Exception
        """
        self.log.debug("Looking for product...")
        products = self.get_product_id_from_errata()
        if len(products) < 1:
            self.log.error("Error: no products found")
            raise RuntimeError(
                "There are no products assigned for erratum '%s'" %
                self.erratum)
        elif len(products) == 1:
            self.log.info("Using product: '%s'", products[0])
            return products[0]
        else:
            self.log.error("Error: multiple products found")
            raise ChoiceException("Erratum '%s' has multiple choices (%s). "
                                  "Use --product-version to choose product version "
                                  "to which you want to add the build." %
                                  (self.erratum, ", ".join(products)))

    def _ensure_newfiles(self):
        """
        Ensure that erratum is in NEW_FILES. This is need in managing build and
        managing bugs.
        """
        self.log.debug("Ensuring that erratum is in NEW_FILES.")
        # if we know that errata is not in new_files, move it
        if self.erratum_json is not None and \
           self.erratum_json['status'] != 'NEW_FILES':
            self.log.debug("Erratum is not in NEW_FILES (state is '%s'), "
                           "moving to NEW_FILES", self.erratum_json['status'])
            self.move_to_newfiles()

        if self.erratum_json is None:
            # if user provides erratum ID, lets try to move it to new_files and
            # ignore error two requests are better than one
            try:
                self.log.debug("Gracefully moving erratum to NEW_FILES")
                self.move_to_newfiles()
            except RuntimeError:
                # TODO: what if not?
                self.log.debug("Erratum was probably in NEW_FILES already, "
                               "this is not an error.")

    def get_product_id_from_errata(self):
        """
        Return all product version names for specified errata.
        """
        self.log.debug("Getting product ID from erratum.")
        adv_js = self.erratum_json or self.et.get_advisory(self.erratum)
        try:
            rel_id = adv_js['group_id']
        except KeyError:
            self.log.error("Error: Can't find advisory's release.")
            raise RuntimeError("Can't find advisory's release.")
        self.log.debug("Release is '%s'", rel_id)

        rel_js = self.et.get_release(rel_id)
        try:
            product_versions = [x['name'] for x in
                                rel_js['data']['relationships']['product_versions']]
        except KeyError:
            self.log.error("Error: no product version found")
            raise RuntimeError("No product versions.")
        else:
            self.log.debug("Product versions found: %s",
                           ", ".join(product_versions))
            return product_versions

    def add_builds(self, nvrs, nowait_rpmdiff):
        """
        add build to erratum

        If nvrs is provided only, there are mechanisms to figure out erratum
        and product:

         * we search for erratum -- we try to find all opened erratas
         * product: if we have erratum, we can get product ID from advisory (it
           has product assigned)

        If we are not sure, we have to let user decide (in case of multiple
        products/erratas matching)

        nvrs [list] -- name-version-release of build built in brew
        """
        self.log.debug("Adding build to erratum.")

        self._erratum_setdefault()
        self._product_setdefault()

        self._ensure_newfiles()

        if not self.dry_log("I would add build '%s' to erratum '%s' and product '%s'",
                            ', '.join(nvrs), self.erratum, ', '.join(self.product_versions)):
            for product_version in self.product_versions:
                try:
                    self.et.add_builds(self.erratum, product_version, nvrs)
                except Exception as ex:
                    self.log.error("Error while adding build: %s", ex)
                    raise RuntimeError("Error while adding build: %s" % ex)

        if not nowait_rpmdiff:
            if not self.dry_log("I would watch rpmdiff task for advisory '%s'",
                                self.erratum_json['fulladvisory']):
                watcher = ETRPMDiffMessageWatcher(self.et_conf, self.umb_config)
                response = watcher.wait_for_erratum(self.erratum_json['errata_id'])
                # ensure response is dict
                if not isinstance(response, dict):
                    self.log.debug("response from rpmdiff.wait_for_erratum = '%s'", response)
                    self.log.error("There was a bad response in processing messages from RPMDiff.")
                if response['status'] == 'OK':
                    return self.move_to_qe()
                elif response['status'] == 'BAD':
                    self.log.error(response['message'])
                else:
                    self.log.debug("response from rpmdiff.wait_for_erratum = '%s'", response)
                    self.log.error("There was an error in processing messages from RPMDiff.")
            else:
                # dry-run: do not fail
                return True
        else:
            if self.dry_log("I would not watch rpmdiff task for advisory '%s'",
                            self.erratum_json['fulladvisory']):
                self.log.warning("Status of erratum '%s' was not changed to QE state. "
                                 "It can be done manually." % (self.erratum))

    def move_to_state(self, state, message=None):
        """Move provided erratum to a specific state"""
        assert state in ('NEW_FILES', 'QE')

        self.log.debug("Moving erratum to %s.", state)
        self._erratum_setdefault()
        dry_run = self.dry_log(
            "I would move erratum '%s' to %s with message \"%s\"",
            self.erratum, state, message)
        if not dry_run:
            try:
                return self.et.change_state(self.erratum, state, message)
            except Exception as ex:
                self.log.error("Couldn't move provided errata %s to %s: %s",
                               self.erratum, state, ex)
                raise RuntimeError(
                    "Couldn't move provided errata {0} to {1}: {2}".format(
                        self.erratum, state, ex))

    def move_to_qe(self, message=None):
        """
        move provided erratum to QE state
        """
        return self.move_to_state('QE')

    def move_to_newfiles(self, message=None):
        """
        move provided erratum to NEW_FILES state
        """
        return self.move_to_state('NEW_FILES')

    def add_bug(self, bug_id):
        """
        Bug is synced first and then it is being added to the advisory

        bug_id [number] -- bugzilla ID
        """
        self.log.debug("Adding bug to erratum.")
        self.log.debug("Forcing ET to sync bug '%s'", bug_id)
        self._erratum_setdefault()
        try:
            self.et.sync_bug(bug_id)
        except RuntimeError as ex:
            self.log.error("Error while syncing bug: %s", ex)
            raise
        self._ensure_newfiles()

        proper_status = ("VERIFIED", "MODIFIED")
        bug_status = self.bz.get_status(bug_id)
        if bug_status not in proper_status:
            message = "Bug '#%s' doesn't have proper status. It has '%s'. Possibilities: %s. " \
                      "Bug can not be added." % (bug_id, bug_status, proper_status)
            self.log.error(message)
            raise RuntimeError(message)

        if not self.dry_log("I would add bug '%s' to erratum '%s'",
                            bug_id, self.erratum):
            try:
                return self.et.add_bug(self.erratum, bug_id)
            except RuntimeError as ex:
                self.log.error("Error while adding bug to erratum: %s", ex)
                raise
        else:
            # dry-run: do not fail
            return True

    def remove_bug(self, bug_id):
        """
        Remove bug from erratum

        bug_id [number] -- bugzilla ID
        """
        self.log.debug("Remove bug from erratum.")
        self._erratum_setdefault()
        self._ensure_newfiles()
        if not self.dry_log("I would remove bug '%s' from erratum '%s'",
                            bug_id, self.erratum):
            try:
                return self.et.remove_bug(self.erratum, bug_id)
            except RuntimeError as ex:
                self.log.error("Error while removing bug from erratum: %s", ex)
                raise


class ErrataWizard(object):
    """
    This is an enhancement of ErrataManager. It is able to run multiple
    commands within one session (as of June 2014, it doesn't make sense much,
    since ET doesn't support sessions and authenticates every request with
    Kerberos, hopefully this will change in future)
    """

    def __init__(self, et_conf, rpmdiff_msg_conf, package=None, erratum=None,
                 product_versions=None, bugs_to_add=None, bugs_to_remove=None,
                 builds=None, move_to_newfiles=None, move_to_qe=None,
                 dry_run=False, log=None, debug=False, verbose=False,
                 username=None, password=None, message=None, nowait_rpmdiff=False,
                 bugzilla_url=None, bugzilla_username=None, bugzilla_password=None):
        """
        you have to specify either package or erratum

        :param dict et_conf: configuration of errata tool containing base
        address
        :param str rpmdiff_msg_conf: configuration of message bus where rpmdiff
        sends updates
        """
        self.package = package
        # options
        self.erratum = erratum
        self.product_versions = product_versions
        self.nowait_rpmdiff = nowait_rpmdiff
        # commands
        self.bugs_to_add = bugs_to_add
        self.bugs_to_remove = bugs_to_remove
        self.builds = builds  # add build to erratum
        self.move_to_newfiles = move_to_newfiles
        self.move_to_qe = move_to_qe
        self.message = message

        self.etm = ErrataManager(et_conf, rpmdiff_msg_conf,
                                 package=package, erratum=erratum,
                                 product_versions=product_versions,
                                 dry_run=dry_run,
                                 log=log, debug=debug, verbose=verbose,
                                 username=username, password=password,
                                 bugzilla_url=bugzilla_url,
                                 bugzilla_username=bugzilla_username,
                                 bugzilla_password=bugzilla_password)

    def run(self):
        """
        run multiple operations at the same time (these may be specified in
        init)
        """
        if self.move_to_newfiles:
            self.etm.move_to_newfiles(self.message)
        if self.bugs_to_remove:
            for bug in self.bugs_to_remove:
                self.etm.remove_bug(bug)
        if self.bugs_to_add:
            for bug in self.bugs_to_add:
                self.etm.add_bug(bug)
        if self.builds:
            self.etm.add_builds(self.builds, self.nowait_rpmdiff)
        if self.move_to_qe:
            self.etm.move_to_qe(self.message)


def test_adv(advisory):
    print(ErrataManager.get_product_id_from_errata(advisory))


if __name__ == '__main__':
    advisory = 17754
    test_adv(advisory)
