# -*- coding: utf-8 -*-

"""
Wrapper on top of curl session.

Copyright (C) 2014 Red Hat Inc.
Author(s): Tomas Tomecek <ttomecek@redhat.com>

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
the full text of the license.
"""

from six import BytesIO, string_types
from six.moves.urllib.parse import urlencode

import pycurl

from .service import LoggingService


class HTTPSession(LoggingService):
    """
    curl will cache http session, so kerberos authentication will happen only
    once
    """

    def __init__(self, debug=False, verbose=False, json=False, username=None,
                 password=None, *args, **kwargs):
        """
        :param username: username for HTTP basic authentication. Will use
                         Kerberos (GSS-API) if not set.
        :param password: password for HTTP basic authentication. See username.
        """
        super(HTTPSession, self).__init__(*args, **kwargs)
        self._c = None
        self.response = BytesIO()
        self.debug = debug
        self.verbose = verbose
        self.json = json
        self.username = username
        self.password = password

    @property
    def c(self):
        if self._c is None:
            self._c = self._init_curl()
        return self._c

    def _init_curl(self):
        curl = pycurl.Curl()
        curl.setopt(pycurl.FOLLOWLOCATION, 1)
        curl.setopt(pycurl.COOKIEFILE, '')
        if self.username and self.password:
            curl.setopt(pycurl.USERPWD, '%s:%s' % (self.username,
                                                   self.password))
        else:
            curl.setopt(pycurl.USERPWD, ':')
            curl.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_GSSNEGOTIATE)
        curl.setopt(pycurl.WRITEFUNCTION, self.response.write)
        curl.setopt(pycurl.SSL_VERIFYPEER, 0)
        curl.setopt(pycurl.VERBOSE, 1 if self.debug else 0)
        if self.json:
            curl.setopt(pycurl.HTTPHEADER, ["Content-Type: application/json",
                                            "Accept: application/json"])
        return curl

    def request(self, url, method, data=None):
        method = method.lower()
        if method == 'post':
            self.c.setopt(pycurl.POST, 1)
        elif method == 'get':
            self.c.setopt(pycurl.HTTPGET, 1)
        else:
            raise RuntimeError(
                "Unsupported method '%s' for curl call!" % method)
        if data:
            is_str = isinstance(data, string_types)
            # 'requests' module do this much better
            if is_str:
                self.c.setopt(pycurl.POSTFIELDS, data)
            else:
                enc_data = urlencode(data, doseq=True)
                self.c.setopt(pycurl.POSTFIELDS, enc_data)

        self.c.setopt(pycurl.URL, url)
        self.log.debug("Performing '%s' request to '%s' with data '%s'",
                       method, url, data)
        self.c.perform()
        response_text = self.response.getvalue().decode("utf-8")
        # clear buffer
        self.response.truncate(0)
        return self.c.getinfo(pycurl.HTTP_CODE), response_text
