# -*- coding: utf-8 -*-
"""
Classes to consume messages via Unified Message Bus (UMB)

Copyright (C) 2017 Red Hat Inc.
Author(s): Chenxiong Qi <cqi@redhat.com>

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
the full text of the license.

Sample messages:

starting event message:

{
  "run_id": 12345,
  "type": "COMPARISON",
  "source": "ERRATA",
  "package_name": "foo",
  "nvr": "foo-1.2-1.el7",
  "baseline": "foo-1.1-1.el7",
  "errata_id": 54321,
  "overall_score": null
}

completed event message:
{
  "run_id": 12345,
  "type": "COMPARISON",
  "source": "ERRATA",
  "package_name": "foo",
  "nvr": "foo-1.2-1.el7",
  "baseline": "foo-1.1-1.el7",
  "errata_id": 54321,
  "overall_score": "Failed"
}

For more detailed information, refer to https://docs.engineering.redhat.com/x/ICOPAg
"""

import json

from rhmsg.activemq.consumer import AMQConsumer

from .service import LoggingService


class UMBWatcher(LoggingService):
    """Base class for concrete UMB watchers"""

    def __init__(self, umb_config, *args, **kwargs):
        super(UMBWatcher, self).__init__(*args, **kwargs)
        self.consumer_queue = umb_config['queue']
        self.umb_consumer = AMQConsumer(umb_config['brokers'],
                                        certificate=umb_config['certificate_file'],
                                        private_key=umb_config['private_key_file'],
                                        trusted_certificates=umb_config['ca_certs'])

    def wait(self, data=None):
        """Start to watch messages

        Depending on how to handle messages defined in concrete subclass,
        you could pass arbitrary data that is required to handle messages.

        :param data: Optional. Passing arbitrary data for handling messages.
            The value of data can be any type, that totally depends on what
            data is required in order to handle messages. Usually, a dict would
            be more flexible to add more data in the future.
        """
        return self.umb_consumer.consume(self.consumer_queue,
                                         callback=self._message_handler,
                                         data=data)

    def _message_handler(self, message, data=None):
        """Internal method to handle a message got from message bus"""
        raise NotImplementedError('Must be implemented in subclass.')


class RPMDiffMessageWatcher(UMBWatcher):
    """Watcher watching messages sent from RPMDiff"""

    def is_starting_event(self, content):
        """Check if message is sent when starting event happens

        As of writing this patch to migrate to UMB, messages sent by RPMDiff
        has no a specific header to indicate to which event a message belongs.
        Field overall_score is the only way to do that. When it is None, the
        message is sent when starting event happens, otherwise, it is completed
        event message.

        After talking with RPMDiff developer, they will add a header in each
        message, then this line of code should be changed by using that new
        header and it could be much clear and easy to understand.
        """
        return content['overall_score'] is None

    def is_result_bad(self, content):
        return content['overall_score'] in ('Failed', 'Needs inspection', 'Waived')

    def is_result_ok(self, content):
        return content['overall_score'] in ('Info', 'Passed')


class ETRPMDiffMessageWatcher(RPMDiffMessageWatcher):
    """Watcher watching RPMDiff messages for specific Errata Tool advisory"""

    def __init__(self, et_config, umb_config, *args, **kwargs):
        super(ETRPMDiffMessageWatcher, self).__init__(umb_config, *args, **kwargs)
        self.et_config = et_config

    def _compose_info_message(self, run_id):
        return "Your RPMDiff run is complete, it passed.\n" \
               "URL: {0}/rpmdiff/show/{1}".format(self.et_config['url'].strip('/'), run_id)

    def _compose_inspect_message(self, run_id):
        return "Your RPMDiff run is complete. It DIDN'T pass and you have to inspect it.\n" \
               "URL: {0}/rpmdiff/show/{1}".format(self.et_config['url'].strip('/'), run_id)

    def _message_handler(self, message, data=None):
        content = json.loads(message.body)
        self.log.debug('Got RPMDiff msg: {0}'.format(message))

        if self.is_starting_event(content):
            self.log.info("RPMDiff job '%s' for '%s' is starting.",
                          content['run_id'], content['errata_id'])
        else:
            errata_id = content['errata_id']
            run_id = content['run_id']
            status = content['overall_score']

            if data['errata_id'] == errata_id:
                if self.is_result_bad(content):
                    return {'raw_message': content, 'status': 'BAD',
                            'message': self._compose_inspect_message(run_id)}
                elif self.is_result_ok(content):
                    return {'raw_message': content, 'status': 'OK',
                            'message': self._compose_info_message(run_id)}
                else:
                    raise RuntimeError("Invalid status '%s' from RPMDiff" % status)

    def wait_for_erratum(self, errata_id):
        return self.wait(data={'errata_id': errata_id})
