# -*- coding: utf-8 -*-

"""
Copyright (C) 2019 Red Hat Inc.

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
the full text of the license.
"""

import datetime
import os

import requests

from .service import LoggingService

# instead of API version 'v1' (or other specific version) it is possible use 'latest'
PRODUCT_PAGES_API_VERSION = 'api/v7'
TIMEOUT = 60
SCHEDULE_TASK_NAME = "RCM: Default rkpkg build target to z-stream tag"


class ProductPagesService(LoggingService):
    """
    Gathers information from Product Pages REST API
    """
    def __init__(self, pp_url, *args, **kwargs):
        super(ProductPagesService, self).__init__(*args, **kwargs)
        if not pp_url:
            raise RuntimeError('Missing URL to Product Pages in configuration.')
        self.api_url = os.path.join(pp_url, PRODUCT_PAGES_API_VERSION)

    def is_released_as_per_ga(self, label):
        """
        Find out whether a release named in parameter 'label' was already released. It looks
        at 'GA' date.
        """
        request_url = os.path.join(self.api_url, 'releases/%s/?fields=id,shortname,ga_date' % label)
        self.log.debug('Making REST API call to Product Pages: %s', request_url)
        try:
            response = requests.get(
                request_url,
                headers=dict(Accept='application/json'),
                timeout=TIMEOUT)
        except requests.exceptions.RequestException as e:
            msg = 'API call to Product Pages failed: %s' % e
            self.log.warn(msg, exc_info=True)
            raise

        if response.ok:
            try:
                json_data = response.json()
                # convert string to datetime
                date_time_obj = datetime.datetime.strptime(json_data['ga_date'], '%Y-%m-%d')
                today = datetime.datetime.today()
            except Exception as e:
                msg = 'Processing results from API call to Product Pages failed: %s' % e
                self.log.warn(msg, exc_info=True)
                raise
            released = today > date_time_obj
            self.log.debug('%s is %s', label, released and 'released' or 'unreleased')
            return {'found': True, 'result': released}

        return {'found': False, 'result': None}

    def is_released_as_per_schedule_task(self, label):
        """
        Finds out whether a release named in parameter 'label' was already released. It lists
        scheduled tasks for the release and selects specific one that RCM team is interested in.
        Then release status is determined based on the date for the task.
        """
        request_url = os.path.join(
            self.api_url,
            'releases/%s/schedule-tasks/?fields=id,name,date_finish' % label)
        self.log.debug('Making REST API call to Product Pages: %s', request_url)
        try:
            response = requests.get(
                request_url,
                headers=dict(Accept='application/json'),
                timeout=TIMEOUT)
        except requests.exceptions.RequestException as e:
            msg = 'API call to Product Pages failed: %s' % e
            self.log.warn(msg, exc_info=True)
            raise

        if response.ok:
            try:
                json_data = response.json()
            except Exception as e:
                msg = 'Processing results from API call to Product Pages failed: %s' % e
                self.log.warn(msg, exc_info=True)
                raise
            pattern = SCHEDULE_TASK_NAME.lower().strip()
            for task in json_data:
                if task['name'].lower().strip() == pattern:
                    # convert string to datetime
                    date_time_obj = datetime.datetime.strptime(task['date_finish'], '%Y-%m-%d')
                    today = datetime.datetime.today()
                    released = today > date_time_obj
                    self.log.debug('%s is %s', label, released and 'released' or 'unreleased')
                    return {'found': True, 'result': released}

        self.log.debug('Specific schedule task for label \'%s\' wasn\'t not found in Product '
                       'Pages' % label)
        return {'found': False, 'result': None}
